import datetime
import importlib
import os
from enum import Enum
from typing import Any, Callable, Dict, Iterator, Optional, Type, cast

from dateutil.parser import parse as parse_date

from redun.hashing import hash_bytes, hash_struct, hash_tag_bytes
from redun.utils import get_func_source, iter_nested_value, pickle_dumps, pickle_loads

MIME_TYPE_PICKLE = "application/python-pickle"


class TypeError(Exception):
    pass


class InvalidValueError(Exception):
    pass


# Python does not provide a type for None, so we compute it here.
NoneType: Type = type(None)


def get_raw_type_name(raw_type: Any) -> str:
    return "{}.{}".format(raw_type.__module__, raw_type.__name__)


class TypeRegistry:
    """
    A registry for redun types and a dispatch for value methods.
    """

    def __init__(self):
        self._type_name2value_type: Dict[str, Type["Value"]] = {}
        self._raw2proxy_type: Dict[Any, Type["ProxyValue"]] = {}

    def get_type_name(self, raw_type: Type) -> str:
        """
        Returns the type name for a python type.
        """
        value_type = self.get_type(raw_type, use_default=False)
        if value_type:
            assert value_type.type_name
            return value_type.type_name

        # All Value classes should be registered at this point.
        assert not issubclass(raw_type, Value)

        return get_raw_type_name(raw_type)

    def parse_type_name(self, type_name: str) -> Type:
        """
        Parse a type name into a python type.
        """
        if type_name == "builtins.NoneType":
            return NoneType

        value_type = self._type_name2value_type.get(type_name)
        if value_type:
            return value_type

        try:
            module_name, raw_type_name = type_name.rsplit(".", 1)
            return importlib.import_module(module_name).__dict__[raw_type_name]
        except (ValueError, ModuleNotFoundError, KeyError):
            raise TypeError('Unable to import type "{}"'.format(type_name))

    def register(self, value_type: Type["Value"]) -> None:
        """
        Register a Value class with the type system.
        """
        if getattr(value_type, "__args__", None):
            # In python 3.6, parameterized types try to register themselves
            # and can conflict with registering the unparameterized type.
            # We do not need them registered, so we ignore them here.
            return

        # Note: getattr is used because Values start registering even before ProxyValue is defined.
        if getattr(value_type, "proxy", False) and getattr(value_type, "type", None):
            # value_type is a proxy for another raw type.
            self._raw2proxy_type[getattr(value_type, "type")] = cast(Type[ProxyValue], value_type)

        assert isinstance(value_type.type_name, str)
        self._type_name2value_type[cast(str, value_type.type_name)] = value_type

    def _get_proxy_type(self, raw_type: type) -> Optional[Type["Value"]]:
        """
        Search through a type's superclasses for a registered ProxyValue.
        """
        for super_raw_type in raw_type.__class__.mro(raw_type):
            proxy_type = self._raw2proxy_type.get(super_raw_type)
            if proxy_type:
                self._raw2proxy_type[raw_type] = proxy_type
                return proxy_type
        return None

    def get_value(self, raw_value: Any) -> "Value":
        """
        Return a Value instance for a raw python type.
        """
        raw_type = type(raw_value)
        proxy_type = self._get_proxy_type(raw_type)
        if proxy_type:
            return proxy_type(raw_value)

        elif isinstance(raw_value, Value):
            return raw_value

        else:
            return ProxyValue(raw_value)

    def get_type(self, raw_type: Type, use_default: bool = True) -> Optional[Type["Value"]]:
        """
        Return a registered Value class for a raw python type.
        """
        proxy_type = self._get_proxy_type(raw_type)
        if proxy_type:
            return proxy_type

        elif isinstance(raw_type, type) and issubclass(raw_type, Value):
            return raw_type

        elif use_default:
            return ProxyValue

        else:
            return None

    def is_valid(self, value: Any) -> bool:
        """
        Returns True if the value is still good to use since being cached.

        For example, if the value was a File reference and the corresponding
        file on the filesystem has since been removed/altered, the File
        reference is no longer valid.
        """
        return self.get_value(value).is_valid()

    def is_valid_nested(self, nested_value: Any) -> bool:
        """
        Returns True if nested value is still good to use since being cached.
        """
        return all(map(self.is_valid, iter_nested_value(nested_value)))

    def get_hash(self, value: Any, data: Optional[bytes] = None) -> str:
        """
        Hashes a value according to the hash method precedence.
        """
        return self.get_value(value).get_hash()

    def get_serialization_format(self, value: Any) -> str:
        """
        Returns mimetype of serialization.
        """
        return self.get_value(value).get_serialization_format()

    def serialize(self, value: Any) -> bytes:
        """
        Serializes the Value into a byte sequence.
        """
        return self.get_value(value).serialize()

    def deserialize(self, type_name: str, data: bytes) -> Any:
        """
        Returns a deserialization of bytes `data` into a new Value.
        """
        try:
            raw_type = self.parse_type_name(type_name)
        except TypeError:
            raise InvalidValueError(f"Unknown type: {type_name}")
        value_type = self.get_type(raw_type)
        assert value_type
        return value_type.deserialize(raw_type, data)

    def preprocess(self, raw_value: Any, preprocess_args: dict) -> Any:
        """
        Preprocess a value before passing to a Task.
        """
        return self.get_value(raw_value).preprocess(preprocess_args)

    def postprocess(self, raw_value: Any, postprocess_args: dict) -> Any:
        """
        Post process a value resulting from a Task.
        """
        return self.get_value(raw_value).postprocess(postprocess_args)

    def parse_arg(self, raw_type: Type, arg: str) -> Any:
        """
        Parse a command-line argument of type value_type.
        """
        value_type = self.get_type(raw_type, use_default=False)
        if value_type:
            return value_type.parse_arg(raw_type, arg)
        else:
            # Try to parse argument using raw_type constructor.
            return raw_type(arg)

    def iter_subvalues(self, raw_value: Any) -> Iterator["Value"]:
        """
        Iterates through the Value's subvalues.
        """
        return self.get_value(raw_value).iter_subvalues()


# Global singleton instance of TypeRegistry.
_type_registry = TypeRegistry()


def get_type_registry() -> TypeRegistry:
    """
    Return the global singleton instance of TypeRegistry.
    """
    return _type_registry


class MetaValue(type):
    """
    Registers Value classes as they are defined.
    """

    def __init__(cls, name, bases, dct, **kwargs):
        # Use cls.__dict__ to look for class attributes defined directly on class
        # and ignore what is set on super classes.
        if not cls.__dict__.get("type_name"):
            if cls.__dict__.get("type"):
                cls.type_name = get_raw_type_name(cls.type)
            else:
                cls.type_name = get_raw_type_name(cls)

        if getattr(cls, "register", True):
            registry = get_type_registry()
            registry.register(cls)


# In Python 3.6, in order to use metaclass for a Generic class we need to
# inherit from GenericMeta.
# https://github.com/python/typing/issues/449
try:
    # Types are ignored here because GenericMeta has been removed in python 3.7+
    from typing import GenericMeta  # type: ignore

    class MetaValue(MetaValue, GenericMeta):  # type: ignore
        def __init__(cls, name, bases, dct, **kwargs):
            super().__init__(name, bases, dct, **kwargs)

except ImportError:
    pass


class Value(metaclass=MetaValue):
    """
    The base class for tracking inputs and outputs in redun, encapsulating both value
    and state, in a hashable and serializable form.

    This class underpins the implementation of memoization of `redun.Task`s. Hashing is used
    as the sole, unique identifier for a value; there is no fallback to deep equality testing.
    Hence, we can hash `Task` inputs to identify repeated invocations and supply the cached
    result instead. Serialization makes it feasible to cache outputs of tasks.

    In addition to handling ordinary values, i.e., data or objects represented in memory, complex
    workflows often require reasoning about state that may be complex or external, such as
    a database or filesystem. Redun accomplishes reasoning over this state by converting it into
    value-like semantics, after which, redun can proceed without differentiating between values
    and state. Specifically, this class provides that abstraction. Therefore, this class offers
    a more complex lifecycle than would be required for ordinary values.

    The default hashing and serialization methods are designed for simple object values. Child
    classes may customize these as needed; see below for the API details.

    We introduce two atypical lifecycle steps that are needed for more complex cases, especially
    those involving values that wrap state: validity checking and pre/post processing.
    The scheduler will arrange for pre- and post-processing hooks to be called on `Value`
    objects around their use in `Task` implementations. This allows users to implement some
    more complex concepts, such as lazy initialization or to temporarily allocate resources.

    See `docs/source/design.md` for more information about "Validity" of values, which is an
    important concept for this class.
    """

    type_name: Optional[str] = None

    def __init__(self, *args, **kwargs):
        pass

    def is_valid(self) -> bool:
        """
        Returns `True` if the value may be used. For ordinary values, this will typically always
        be true. For state-tracking values, the deserialized cache value may no longer be valid.
        If this method returns `False`, the object will typically be discarded.

        This method may be called repeatedly on the same object.

        For example, if the value was a `File` reference and the corresponding
        file on the filesystem has since been removed/altered, the `File`
        reference is no longer valid.
        """
        return True

    def get_hash(self, data: Optional[bytes] = None) -> str:
        """
        Returns a hash for the value. Users may override to provide custom behavior.

        Most basic objects are covered by the default approach of hashing the binary
        pickle serialization.
        """
        if data is None:
            data = pickle_dumps(self)
        return hash_tag_bytes("Value", data)

    def get_serialization_format(self) -> str:
        """
        Returns mimetype of serialization.
        """
        return MIME_TYPE_PICKLE

    def serialize(self) -> bytes:
        """
        Serializes the Value into a byte sequence.
        """
        return pickle_dumps(self)

    @classmethod
    def deserialize(cls, raw_type: type, data: bytes) -> Any:
        """
        Returns a deserialization of bytes `data` into a new Value.
        """
        return pickle_loads(data)

    def __getstate__(self) -> dict:
        """
        Returns a plain python datastructure for serialization.

        This is the standard pickle method, which users may implement as needed.
        """
        raise NotImplementedError()

    def __setstate__(self, state: dict) -> None:
        """
        Populates the value from state during deserialization.

        This is the standard pickle method, which users may implement as needed.
        """
        raise NotImplementedError()

    def preprocess(self, preprocess_args: dict) -> "Value":
        """
        Preprocess a value before passing it to a Task, and return the preprocessed value.

        This is a scheduler lifecycle method and is not typically invoked by users.
        This method may be called repeatedly on a particular instance, for example, if a single
        `Value` is passed to multiple `Task`s.

        Parameters
        ----------
        preprocess_args : dict
            Extra data provided by the scheduler. Implementations that are not part of redun
            should discard this argument.
        """
        # Ordinary values don't need to do anything.
        return self

    def postprocess(self, postprocess_args: dict) -> "Value":
        """
        Post process a value resulting from a Task, and return the postprocessed value.

        This is a scheduler lifecycle method and is not typically invoked by users.
        This method may be called repeatedly on a particular instance, for example, if it is
        returned recursively.

        Parameters
        ----------
        postprocess_args : dict
            Extra data provided by the scheduler. Implementations that are not part of redun
            should discard this argument.
        """
        # Ordinary values don't need to do anything.
        return self

    @classmethod
    def parse_arg(cls, raw_type: type, arg: str) -> Any:
        """
        Parse a command line argument in a new Value.
        """
        # By default, try to use constructor to parse command line argument.
        return cls(arg)

    def iter_subvalues(self) -> Iterator["Value"]:
        """
        Iterates through the Value's subvalues.
        """
        if False:
            # By default, Values have no subvalues, but we need the yield
            # statement here to convert this method into an iterator.
            # Subclasses can then override this method for their own needs.
            yield
        return


class ProxyValue(Value):
    """
    Class for overriding Value behavior (hashing, serialization) for raw python types.
    """

    proxy = True
    type: Any = None

    def __init__(self, instance: Any):
        self.instance = instance

    def get_hash(self, data: Optional[bytes] = None) -> str:
        """
        Returns a hash for the value.
        """
        if data is None:
            data = pickle_dumps(self.instance)
        return hash_tag_bytes("Value", data)

    def serialize(self) -> bytes:
        """
        Serializes the Value into a byte sequence.
        """
        return pickle_dumps(self.instance)

    def preprocess(self, preprocess_args) -> "Value":
        """
        Preprocess a value before passing to a Task.
        """
        return self.instance

    def postprocess(self, postprocess_args) -> "Value":
        """
        Post process a value resulting from a Task.
        """
        return self.instance

    def iter_subvalues(self) -> Iterator["Value"]:
        """
        Iterates through the Value's subvalues.
        """
        if isinstance(self.instance, Value):
            # No subvalues.
            return
        for subvalue in iter_nested_value(self.instance):
            if isinstance(subvalue, Value):
                yield subvalue
                yield from subvalue.iter_subvalues()


class Bool(ProxyValue):
    """
    Augment builtins.bool to support argument parsing.
    """

    type = bool
    type_name = "builtins.bool"

    @classmethod
    def parse_arg(cls, raw_type: Type, arg: str) -> Any:
        arg = arg.lower()
        if arg == "true":
            return True
        elif arg == "false":
            return False
        else:
            raise ValueError('Unknown bool "{}"'.format(arg))


class Set(ProxyValue):
    """
    Augment builtins.set to support stable hashing.
    """

    type = set
    type_name = "builtins.set"

    def get_hash(self, data: Optional[bytes] = None) -> str:
        # Sort the set to ensure stable serialization and hashing.
        bytes = pickle_dumps(sorted(self.instance))

        # Use a unique tag to distinguish from hashing a list.
        return hash_tag_bytes("Value.set", bytes)


class EnumType(ProxyValue):
    """
    Augment enum.Enum to support argument parsing with choices.
    """

    type = Enum
    type_name = "enum.Enum"

    @classmethod
    def parse_arg(cls, raw_type: Type, arg: str) -> Any:
        name2item = {item.name: item for item in raw_type}
        try:
            if "." in arg:
                enum_name, arg = arg.split(".", 1)
                if enum_name != raw_type.__name__:
                    raise KeyError
            return name2item[arg]
        except KeyError:
            raise ValueError("{} is not a valid {}".format(arg, raw_type.__name__))


class DatetimeType(ProxyValue):
    """
    Augment datetime.datetime to support argument parsing.
    """

    type = datetime.datetime
    type_name = "datetime.datetime"

    @classmethod
    def parse_arg(cls, raw_type: Type, arg: str) -> datetime.datetime:
        return parse_date(arg)


# Get the type for python functions.
function_type = type(lambda: None)


def make_unknown_function(func_name: str) -> Callable:
    """
    Returns a stub function in place of a function that could not be reimported.
    """

    def unknown_function(*args, **kwargs):
        raise ValueError(f"Function '{func_name}' cannot be found.")

    # Mark this function as an unknown function.
    unknown_function.unknown_function = True  # type: ignore
    return unknown_function


def is_unknown_function(func: Callable) -> bool:
    """
    Returns True if the function was unknown when trying to reimport.
    """
    return getattr(func, "unknown_function", False)


class Function(ProxyValue):
    """
    Value class to allow redun to hash and cache plain Python functions.
    """

    type = function_type
    type_name = "builtins.function"

    def __init__(self, instance: Callable):
        super().__init__(instance)
        self.module = instance.__module__
        self.name = instance.__name__
        self._hash = self._calc_hash()

        if not is_unknown_function(instance) and "<locals>" in instance.__qualname__:
            raise InvalidValueError(
                "functions used as redun arguments or results must be defined globally."
            )

    def __repr__(self) -> str:
        return f"<function {self.fullname}>"

    @property
    def fullname(self) -> str:
        """
        Return a fullname for a function including its module.
        """
        return f"{self.module}.{self.name}"

    def _calc_hash(self) -> str:
        # Hash based on fully qualified function name and source code.
        hash = hash_struct(
            [
                "Value.function",
                self.fullname.encode("utf8"),
                "source",
                get_func_source(self.instance),
            ]
        )
        return hash

    def __getstate__(self) -> dict:
        return {
            "hash": self._hash,
            "module": self.module,
            "name": self.name,
        }

    def __setstate__(self, state: dict) -> None:
        self._hash = state["hash"]
        self.module = state["module"]
        self.name = state["name"]
        try:
            # Try to reassociate the function based on its name.
            module = importlib.import_module(self.module)
            self.instance = getattr(module, self.name)
        except (ModuleNotFoundError, AttributeError):
            self.instance = make_unknown_function(self.fullname)

    def get_hash(self, data: Optional[bytes] = None) -> str:
        return self._hash

    def is_valid(self) -> bool:
        """
        Value is valid to use from the cache if we can reassociate the function.
        """
        return not is_unknown_function(self.instance)

    def serialize(self) -> bytes:
        # Serialize the Function wrapper.
        return pickle_dumps(self)

    @classmethod
    def deserialize(self, raw_type: Any, data: bytes) -> Any:
        # Deserialize the Function wrapper and return the inner python function.
        func_wrapper = pickle_loads(data)
        return func_wrapper.instance

    @classmethod
    def parse_arg(cls, raw_type: Type, arg: str) -> Any:
        """
        Parses function by fully qualified name from command-line.
        """
        try:
            module_name, func_name = arg.rsplit(".", 1)
        except ValueError:
            raise ValueError(f"Unexpected format for function name: {arg}")

        try:
            module = importlib.import_module(module_name)
            func = getattr(module, func_name)
        except (ModuleNotFoundError, AttributeError):
            raise ValueError(f"Function not found: {arg}")

        return func


class FileCache(ProxyValue):
    """
    Baseclass for caching values to Files.

    This baseclass can be used to control where the serialized data of
    potentially large values is stored.

    .. code-block:: python

        # Example of user-specific data type.
        class Data:
            def __init__(self, data):
                self.data = data

        # Subclassing from FileCache and setting `type = Data` will register
        # with redun that File-caching should be used for any Value of type
        # `Data`.
        class DataType(FileCache):
            type = Data

            # This path specifies where the files containing serialized
            # values should be stored. This variable accepts all paths
            # acceptable to File (e.g. local, s3, etc).
            base_path = "tmp"

        @task()
        def task1() -> Data:
            # This value will be serialized to a File instead of the default
            # redun database.
            return Data("my data")
    """

    base_path = "."

    def _serialize(self) -> bytes:
        # User defined serialization.
        return pickle_dumps(self.instance)

    @classmethod
    def _deserialize(cls, data: bytes) -> Any:
        # User defined deserialization.
        return pickle_loads(data)

    def serialize(self) -> bytes:
        from redun.file import File

        # Serialize data.
        bytes = self._serialize()

        # Write (possibly large) data to file.
        filename = os.path.join(self.base_path, hash_bytes(bytes))
        File(filename).write(bytes, mode="wb")

        # Only cache filename.
        return filename.encode("utf8")

    @classmethod
    def deserialize(cls, raw_type: type, filename: bytes) -> Any:
        from redun.file import File

        # Read data from file.
        file = File(filename.decode("utf8"))
        if not file.exists():
            # File has since been delete. Treat as cache miss.
            raise InvalidValueError()

        data = cast(bytes, file.read("rb"))
        return cls._deserialize(data)



import datetime
import importlib
import inspect
import linecache
import logging
import os
import pprint
import queue
import shlex
import sys
import threading
import time
import traceback
import uuid
from collections import OrderedDict, defaultdict
from itertools import chain, takewhile
from traceback import FrameSummary
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from redun.backends.base import RedunBackend, TagEntity
from redun.backends.db import RedunBackendDb
from redun.config import Config, Section, SectionProxy, create_config_section
from redun.executors.base import Executor, get_executors_from_config
from redun.executors.local import LocalExecutor
from redun.expression import (
    AnyExpression,
    ApplyExpression,
    Expression,
    QuotedExpression,
    SchedulerExpression,
    SimpleExpression,
    TaskExpression,
    ValueExpression,
    derive_expression,
    get_lazy_operation,
    quote,
)
from redun.handle import Handle
from redun.hashing import hash_eval, hash_struct
from redun.logging import logger as _logger
from redun.promise import Promise, wait_promises
from redun.tags import parse_tag_value
from redun.task import (
    CacheCheckValid,
    CacheResult,
    CacheScope,
    Task,
    TaskRegistry,
    get_task_registry,
    scheduler_task,
    task,
)
from redun.utils import format_table, iter_nested_value, map_nested_value, trim_string
from redun.value import Value, get_type_registry

# Globals.
_local = threading.local()

# Constants.
JOB_ACTION_WIDTH = 6  # Width of job action in logs.

Result = TypeVar("Result")
T = TypeVar("T")
S = TypeVar("S")


def settrace_patch(tracefunc: Any) -> None:
    """
    Monkey patch for recording whether debugger REPL is active or not.

    sys.settrace() is called by debuggers such as pdb whenever the debugger
    REPL is activated. We use a monkey patch in order to record tracing
    process-wide. Relying on `sys.gettrace()` is not enough because it is
    thread-specific.
    """
    global _is_debugger_active
    _is_debugger_active = bool(tracefunc)
    try:
        _original_settrace(tracefunc)
    except Exception:
        # IDEs, such as PyCharm, may ban calls to settrace().
        # http://pydev.blogspot.com/2007/06/why-cant-pydev-debugger-work-with.html
        # In such cases, do nothing.
        pass


def is_debugger_active() -> bool:
    """
    Returns True if debugger REPL is currently active.
    """
    global _is_debugger_active
    return _is_debugger_active


# Patch sys.settrace() in order to detect presence of debugger.
_original_settrace = sys.settrace
_is_debugger_active = False
sys.settrace = settrace_patch


class NoCurrentScheduler(Exception):
    def __init__(self):
        super().__init__("Scheduler is not running.")


class SchedulerError(Exception):
    pass


class DryRunResult(Exception):
    pass


def get_current_scheduler(required=True) -> Optional["Scheduler"]:
    """
    Returns the currently running Scheduler for this thread.
    """
    if not hasattr(_local, "_redun_scheduler"):
        if required:
            raise NoCurrentScheduler()
        else:
            return None
    return _local._redun_scheduler


def set_current_scheduler(scheduler: Optional["Scheduler"]) -> None:
    """
    Sets the current running Scheduler for this thread.
    """
    if scheduler:
        _local._redun_scheduler = scheduler
    else:
        try:
            del _local._redun_scheduler
        except AttributeError:
            pass


def set_arg_defaults(task: "Task", args: Tuple, kwargs: dict) -> Tuple[Tuple, dict]:
    """
    Set default arguments from Task signature.
    """
    # Start with given kwargs.
    kwargs2 = dict(kwargs)

    sig = task.signature
    for i, param in enumerate(sig.parameters.values()):
        if i < len(args):
            # User already specified this arg in args.
            continue

        elif param.name in kwargs2:
            # User already specified this arg in kwargs.
            continue

        elif param.default is not param.empty:
            # Default should be used.
            kwargs2[param.name] = param.default
    return args, kwargs2


def format_arg(arg_name: str, value: Any, max_length: int = 200) -> str:
    """
    Format a Task argument into a string.
    """
    return "{arg_name}={value}".format(
        arg_name=arg_name, value=trim_string(repr(value), max_length=max_length)
    )


def format_task_call(task: "Task", args: Tuple, kwargs: dict) -> str:
    """
    Format a Task call into a string.

    ```
    my_task(arg1=10, my_file=File('path/to/file.txt')
    ```
    """
    all_args = OrderedDict()
    sig = task.signature

    for i, param in enumerate(sig.parameters.values()):
        if i < len(args):
            # Positional argument.
            all_args[param.name] = args[i]

        else:
            # Keyword argument.
            all_args[param.name] = kwargs.get(param.name, param.default)

    args_text = ", ".join(format_arg(arg_name, value) for arg_name, value in all_args.items())
    return "{task}({args})".format(
        task=task.fullname,
        args=args_text,
    )


class Execution:
    """
    An Execution tracks the execution of a workflow of :class:`redun.task.Task`s.
    """

    def __init__(self, id: str):
        self.id = id
        self.job: Optional[Job] = None

    def add_job(self, job: "Job") -> None:
        # Record first job as root job.
        if not self.job:
            self.job = job


class Job:
    """
    A Job tracks the execution of a :class:`redun.task.Task` through its various stages.
    """

    STATUSES = ["PENDING", "RUNNING", "FAILED", "CACHED", "DONE", "TOTAL"]

    def __init__(
        self,
        task: Task,
        expr: TaskExpression,
        id: Optional[str] = None,
        parent_job: Optional["Job"] = None,
        execution: Optional[Execution] = None,
    ):
        self.id = id or str(uuid.uuid4())

        # The Task used in expr.
        self.task = task

        # TaskExpression for which this job was created to evaluate.
        self.expr = expr

        # Execution representing the entire running workflow.
        self.execution: Optional[Execution] = execution

        # Execution state.

        # Job-level task option overrides.
        self.task_options: Dict[str, Any] = {}

        # The evaluated version of (self.expr.args, self.expr.kwargs)
        self.eval_args: Optional[Tuple[Tuple, Dict]] = None

        # The evaluated and preprocessed arguments.
        # These are the arguments (args, kwargs) submitted to executors.
        self.args: Optional[Tuple[Tuple, Dict]] = None

        # The hash of the evaluated and preprocessed arguments.
        self.args_hash: Optional[str] = None

        # The hash of (self.task.hash, self.arg_hash). This is used as the cache key.
        self.eval_hash: Optional[str] = None

        # This is true if we fetched the result from the cache.
        self.was_cached: bool = False

        # Hash of the CallNode associated with running this job. This hash requires knowledge
        # of the Job's result, hence is available after either computing or retrieving the result.
        self.call_hash: Optional[str] = None

        # Promise for evaluating self.expr.
        self.result_promise: Promise = Promise()

        # The child jobs created for task calls within self.expr.
        self.child_jobs: List[Job] = []

        # The parent job or None if this is the root job of the execution.
        self.parent_job: Optional[Job] = parent_job

        # Bookkeeping for handles created during this job.
        self.handle_forks: Dict[str, int] = defaultdict(int)

        # Tags (key-value pairs) that should be added to this job.
        self.job_tags: List[Tuple[str, Any]] = []

        # Tags (key-value pairs) that should be added to the execution.
        self.execution_tags: List[Tuple[str, Any]] = []

        # Tags (value_hash and key-value pairs) that should be added to values.
        self.value_tags: List[Tuple[str, List[Tuple[str, Any]]]] = []

        # A cached status of the job (PENDING, RUNNING, CACHED, DONE, FAILED).
        self._status: Optional[str] = None

        if parent_job:
            self.add_parent(parent_job)
        if execution:
            execution.add_job(self)

    def __repr__(self) -> str:
        return f"Job(id={self.id}, task_name={self.task.fullname})"

    @property
    def status(self) -> str:
        if self._status:
            return self._status
        elif self.eval_args is None:
            return "PENDING"
        elif self.result_promise.is_pending:
            return "RUNNING"
        elif self.result_promise.is_fulfilled:
            if self.was_cached:
                return "CACHED"
            else:
                return "DONE"
        elif self.result_promise.is_rejected:
            return "FAILED"
        else:
            raise ValueError("Unknown status")

    def get_option(self, key: str, default: Any = None, as_type: Optional[Type] = None) -> Any:
        """
        Returns a task option associated with a :class:`Job`.

        Precedence is given to task options defined at call-time
        (e.g. `task.options(option=foo)(arg1, arg2)`) over task definition-time
        (e.g. `@task(option=foo)`).
        """
        assert "task_expr_options" in self.expr.__dict__
        task = cast(Task, self.task)

        def get_raw():
            if key in self.task_options:
                return self.task_options[key]
            elif key in self.expr.task_expr_options:
                return self.expr.task_expr_options[key]
            elif task.has_task_option(key):
                return task.get_task_option(key)
            else:
                return default

        result = get_raw()
        return as_type(result) if as_type else result

    def get_options(self) -> dict:
        """
        Returns task options for this job.

        Precedence is given to task options defined at call-time
        (e.g. `task.options(option=foo)(arg1, arg2)`) over task definition-time
        (e.g. `@task(option=foo)`).
        """
        assert self.task
        task_options = {
            **self.task.get_task_options(),
            **self.expr.task_expr_options,
            **self.task_options,
        }
        return task_options

    def get_limits(self) -> Dict[str, int]:
        """
        Returns resource limits required for this job to run.
        """
        limits = self.get_option("limits", {}) if self.task else {}

        # We allow limits to be a list of resources. In that case, we default the required
        # resource count to 1 for each resource specified. We create the mapping of limit to
        # count here so that we always have a dict when constructing job_limits below.
        if isinstance(limits, list):
            limits = {limit_name: 1 for limit_name in limits}

        assert isinstance(limits, dict)

        job_limits: Dict[str, int] = defaultdict(int)
        job_limits.update(limits)
        return job_limits

    def add_parent(self, parent_job: "Job") -> None:
        """
        Maintains the Job tree but connecting the job with a parent job.
        """
        parent_job.child_jobs.append(self)

    def collapse(self, other_job: "Job") -> None:
        """
        Collapse this Job into `other_job`.

        This method is used when equivalent Jobs are detected, and we want to
        perform Common Subexpression Elimination (CSE).
        """

        # Inform parent job, to use other_job instead.
        parent_job = self.parent_job
        assert parent_job
        parent_job.child_jobs[parent_job.child_jobs.index(self)] = other_job

        # Make callbacks just as if we had gotten a cache hit.
        def then(result: Any) -> None:
            self.call_hash = other_job.call_hash
            scheduler = get_current_scheduler()
            assert scheduler
            self.was_cached = True

            scheduler.done_job(self, result)

        def fail(error: Any) -> None:
            self.call_hash = other_job.call_hash
            scheduler = get_current_scheduler()
            assert scheduler
            assert other_job.eval_args
            self.was_cached = True

            # It's important to use the main thread directly, because if we queue up the
            # event, the scheduler may shut down before processing it. This way, we can ensure
            # all the bookkeeping gets done synchronously.
            scheduler._reject_job_main_thread(self, error)

        other_job.result_promise.then(then, fail)

    def resolve(self, result: Any) -> None:
        """
        Resolves a Job with a final concrete value, `result`.
        """
        self.expr.call_hash = self.call_hash
        self.result_promise.do_resolve(result)
        self.clear()

    def reject(self, error: Any) -> None:
        """
        Rejects a Job with an error exception.
        """
        self.expr.call_hash = self.call_hash
        self.result_promise.do_reject(error)
        self.clear()

    def clear(self):
        """
        Free execution state from Job.
        """
        # Record final status before clearing execution state.
        self._status = self.status

        # Implementation note: we need to do this because the job holds
        # both parent and child jobs, creating a reference cycle that would
        # prevent garbage collection. Here, we break that cycle.
        self.expr = None
        self.eval_args = None
        self.args = None
        self.result_promise = None
        self.job_tags.clear()
        self.execution_tags.clear()
        self.value_tags.clear()
        self.child_jobs.clear()


def get_backend_from_config(backend_config: Optional[SectionProxy] = None) -> RedunBackend:
    """
    Parses a redun :class:`redun.backends.base.RedunBackend` from a config object.
    """
    if not backend_config:
        backend_config = create_config_section()
    backend_config = cast(SectionProxy, backend_config)
    if not backend_config.get("db_uri"):
        # By default, use in-memory db and autoload (create schemas).
        backend_config["db_uri"] = "sqlite:///:memory:"
        load = True
    else:
        load = False

    backend = RedunBackendDb(config=backend_config)
    if load:
        backend.load()
    return backend


def get_limits_from_config(limits_config: Optional[Section] = None) -> Dict[str, int]:
    """
    Parses resource limits from a config object.
    """
    return (
        {key: int(value) for key, value in cast(dict, limits_config).items()}
        if limits_config
        else {}
    )


def get_ignore_warnings_from_config(scheduler_config: Section) -> Set[str]:
    """
    Parses ignore warnings from config.
    """
    warnings_text = scheduler_config.get("ignore_warnings")
    if not warnings_text:
        return set()
    return set(warnings_text.strip().split())


def format_job_statuses(
    job_status_counts: Dict[str, Dict[str, int]],
    timestamp: Optional[datetime.datetime] = None,
) -> Iterator[str]:
    """
    Format job status table (Dict[task_name, Dict[status, count]]).
    """
    # Create counts table.
    task_names = sorted(job_status_counts.keys())
    table: List[List[str]] = (
        [["TASK"] + Job.STATUSES]
        + [
            ["ALL"]
            + [
                str(sum(job_status_counts[task_name][status] for task_name in task_names))
                for status in Job.STATUSES
            ]
        ]
        + [
            [task_name] + [str(job_status_counts[task_name][status]) for status in Job.STATUSES]
            for task_name in task_names
        ]
    )

    # Display job status table.
    if timestamp is None:
        timestamp = datetime.datetime.now()
    yield "| JOB STATUS {}".format(timestamp.strftime("%Y/%m/%d %H:%M:%S"))

    for line in format_table(table, "lrrrrrr", min_width=7):
        yield f"| {line}"
    yield ""


class Frame(FrameSummary, Value):
    """
    Frame of a :class:`Traceback` for :class:`Job` failure.
    """

    type_name = "redun.Frame"

    def __init__(
        self,
        filename: str,
        lineno: int,
        name: str,
        locals: Dict[str, Any],
        lookup_line: bool = True,
        line: Optional[str] = None,
        job: Optional[Job] = None,
    ):
        self.filename = filename
        self.lineno = lineno
        self.name = name
        self._line = line
        if lookup_line:
            self.line
        self.locals = {key: trim_string(repr(value)) for key, value in locals.items()}
        assert job
        self.job: Job = job

        # Advance past decorator.
        if self._line and self._line.strip().startswith("@"):
            while True:
                self.lineno += 1
                line = linecache.getline(self.filename, self.lineno).strip()
                if line.startswith("def "):
                    break
            self._line = None

    def __getstate__(self) -> dict:
        return {
            "filename": self.filename,
            "lineno": self.lineno,
            "name": self.name,
            "_line": self._line,
            "locals": self.locals,
            "job": self.job,
        }

    def __setstate__(self, state: dict) -> None:
        self.filename = state["filename"]
        self.lineno = state["lineno"]
        self.name = state["name"]
        self._line = state["_line"]
        self.locals = state["locals"]
        self.job = state["job"]


class Traceback(Value):
    """
    Traceback for :class:`Job` failure.
    """

    type_name = "redun.Traceback"

    def __init__(self, error: Any, frames: List[FrameSummary], logs: Optional[List[str]] = None):
        self.error: Any = error
        self.frames = frames
        self.logs: List[str] = logs or []

    def __getstate__(self) -> dict:
        return {
            "error": self.error,
            "frames": self.frames,
            "logs": self.logs,
        }

    def __setstate__(self, state: dict) -> None:
        self.error = state["error"]
        self.frames = state["frames"]
        self.logs = state["logs"]

    @classmethod
    def from_error(self, error: Exception, trim_frames=True) -> "Traceback":
        """
        Returns a new :class:`Traceback` derived from an Exception.

        Parameters
        ----------
        error : Exception
            Exception object from which to derive a Traceback.
        trim_frames : bool
            If True, do not include redun scheduler related frames in Traceback.
        """
        frames = list(traceback.extract_tb(error.__traceback__))
        if trim_frames:
            frames = self.trim_frames(frames)
        return Traceback(error, frames)

    @classmethod
    def trim_frames(self, frames: List[FrameSummary]) -> List[FrameSummary]:
        # As we walk back through the call stack, when we reach a frame from
        # redun's executors or cli, we know we have left the user's code.
        redun_path_prefixes = [
            os.path.join(os.path.dirname(__file__), "executors"),
            os.path.join(os.path.dirname(__file__), "cli.py"),
        ]

        return list(
            reversed(
                list(
                    takewhile(
                        lambda frame: not any(
                            frame.filename.startswith(prefix) for prefix in redun_path_prefixes
                        ),
                        reversed(frames),
                    )
                )
            )
        )

    def format(self) -> Iterator[str]:
        """
        Iterates through lines displaying the :class:`Traceback`.
        """
        for frame in self.frames:
            if isinstance(frame, Frame):
                yield '  Job {job}: File "{file}", line {lineno}, in {task}\n'.format(
                    job=frame.job.id[:8],
                    file=frame.filename,
                    lineno=frame.lineno,
                    task=frame.name,
                )
            else:
                yield '  File "{file}", line {lineno}, in {func}\n'.format(
                    file=frame.filename,
                    lineno=frame.lineno,
                    func=frame.name,
                )
            yield "    {line}\n".format(line=frame.line)
            if frame.locals:
                var_width = max(map(len, frame.locals.keys()))
                for key, value in sorted(frame.locals.items()):
                    yield "    {key} = {value}\n".format(key=key.ljust(var_width), value=value)

        yield "{}: {}".format(type(self.error).__name__, self.error)

        if self.logs:
            yield "\n"
            yield "Latest logs:\n"
            for line in self.logs:
                yield line


class ErrorValue(Value):
    """
    Value for wrapping Exceptions raised by Task.
    """

    type_name = "redun.ErrorValue"

    def __init__(self, error: Exception, traceback: Optional[Traceback] = None):
        assert isinstance(error, Exception)
        self.error = error
        self.traceback = traceback

    def __repr__(self) -> str:
        return "ErrorValue({})".format(repr(self.error))

    def get_hash(self, data: Optional[bytes] = None) -> str:
        registry = get_type_registry()
        return hash_struct(
            ["redun.ErrorValue", registry.get_hash(self.error), registry.get_hash(self.traceback)]
        )

    def __getstate__(self) -> dict:
        return {"error": self.error, "traceback": self.traceback}

    def __setstate__(self, state: dict) -> None:
        self.error = state["error"]
        self.traceback = state["traceback"]


@task(name="root_task", namespace="redun")
def root_task(expr: QuotedExpression[Result]) -> Result:
    """
    Default task used for a root job in an Execution.
    """
    return expr.eval()


def needs_root_task(task_registry: TaskRegistry, expr: Any) -> bool:
    """
    Returns True if expression `expr` needs to be wrapped in a root task.
    """
    # A root expr must be a TaskExpression.
    if not isinstance(expr, TaskExpression) or isinstance(expr, SchedulerExpression):
        return True

    # All arguments must be concrete.
    task = task_registry.get(expr.task_name)
    assert task
    args, kwargs = set_arg_defaults(task, expr.args, expr.kwargs)
    return any(isinstance(arg, Expression) for arg in iter_nested_value((args, kwargs)))


class Scheduler:
    """
    Scheduler for evaluating redun tasks.

    A thread may only have a single running scheduler at a time, and a scheduler can perform
    exactly one execution at a time. While running, the scheduler will register itself into a
    thread-local variable, see `get_current_scheduler` and `set_current_scheduler`.

    A scheduler may be reused for multiple executions, and makes every effort to be stateless
    between executions. That is, the scheduler clears its internal state before starting a new
    execution, providing isolation from any others we may have performed.

    Although the scheduler collaborates with executors that may use multiple threads during
    execution, the scheduler logic relies upon being executed from a single thread. Therefore,
    the main lifecycle methods used by executors defer back to the scheduler thread.
    The scheduler is implemented around an event loop and asynchronous callbacks, allowing it to
    coordinate many in-flight jobs in parallel.

    Scheduler tasks are generally considered "friends" of the scheduler and may need to access
    some private methods.
    """

    def __init__(
        self,
        config: Optional[Config] = None,
        backend: Optional[RedunBackend] = None,
        executor: Optional[Executor] = None,
        logger: Optional[Any] = None,
        use_task_traceback: bool = True,
        job_status_interval: Optional[int] = None,
        migrate: Optional[bool] = None,
    ):
        self.config = config or Config()
        self.logger = logger or _logger
        self.ignore_warnings = get_ignore_warnings_from_config(self.config.get("scheduler", {}))
        self.use_task_traceback = use_task_traceback
        self.traceback: Optional[Traceback] = None
        self.job_status_interval: Optional[int] = job_status_interval or int(
            self.config.get("scheduler", {}).get("job_status_interval", 20)
        )
        if self.job_status_interval <= 0:
            self.job_status_interval = None

        # Setup backend.
        if backend is None:
            backend = get_backend_from_config(self.config.get("backend"))

        self.backend: RedunBackend = backend

        # Setup executors.
        self.executors: Dict[str, Executor] = {}
        if executor:
            self.add_executor(executor)
        else:
            # Add default executors.
            self.add_executor(LocalExecutor("default", mode="thread"))
            self.add_executor(LocalExecutor("process", mode="process"))
        for executor in get_executors_from_config(self.config.get("executors", {})):
            self.add_executor(executor)

        # Setup limits.
        self.limits: Dict[str, int] = get_limits_from_config(self.config.get("limits"))
        self.limits_used: Dict[str, int] = defaultdict(int)
        self._jobs_pending_limits: List[Tuple[Job, Tuple[tuple, dict]]] = []

        # Setup execution tags.
        tags = self.config.get("tags", {})
        self._exec_tags = [(key, parse_tag_value(value)) for key, value in tags.items()]

        # Registries (aka environment).
        self.task_registry = get_task_registry()
        self.type_registry = get_type_registry()

        # Scheduler state.
        self.thread_id: Optional[int] = None
        self.events_queue: queue.Queue = queue.Queue()
        self.workflow_promise: Optional[Promise] = None
        self._current_execution: Optional[Execution] = None

        # Tracking for expression cycles. We store pending expressions by their parent_job
        # because the parent_job's lifetime corresponds to the expression's lifetime.
        # This makes for straightforward deallocation.
        self._pending_expr: Dict[
            Optional[Job], Dict[str, Tuple[Promise, Expression]]
        ] = defaultdict(dict)

        # Tracking for Common Subexpression Elimination (CSE) for pending Jobs.
        self._pending_jobs: Dict[Optional[str], Job] = {}

        # Currently pending/running jobs.
        self._jobs: Set[Job] = set()

        # Job status of finalized jobs.
        self._finalized_jobs: Dict[str, Dict[str, int]] = defaultdict(lambda: defaultdict(int))

        # Execution modes.
        self._dryrun = False
        self._use_cache = True

    @property
    def is_running(self) -> bool:
        return self._current_execution is not None

    def clear(self):
        """Release resources"""
        self._pending_expr.clear()
        self._pending_jobs.clear()
        self._jobs.clear()
        self._finalized_jobs.clear()

    def add_executor(self, executor: Executor) -> None:
        """
        Add executor to scheduler.
        """
        self.executors[executor.name] = executor
        executor.set_scheduler(self)

    def load(self, migrate: Optional[bool] = None) -> None:
        self.backend.load(migrate=migrate)

    def log(
        self, *messages: Any, indent: int = 0, multiline: bool = False, level: int = logging.INFO
    ) -> None:
        text = " ".join(map(str, messages))
        if not multiline:
            lines = text.split("\n")
        else:
            lines = [text]
        for line in lines:
            self.logger.log(level, (" " * indent) + line)

    def _validate_tasks(self) -> None:
        """
        Validate that tasks are properly configured.
        """
        invalid_tasks = [task for task in self.task_registry if not task.namespace]
        if invalid_tasks and "namespace" not in self.ignore_warnings:
            task_names = [
                "{}:{}".format(
                    os.path.basename(task.func.__code__.co_filename), task.func.__name__
                )
                for task in invalid_tasks
            ]
            self.logger.warning(
                "Tasks will require namespace soon. Either set namespace in the `@task` decorator "
                "or with the module-level variable `redun_namespace`.\n"
                "tasks needing namespace: {}".format(", ".join(task_names))
            )

    def _run(
        self,
        expr: TaskExpression[Result],
        parent_job: Optional[Job] = None,
        dryrun: bool = False,
        cache: bool = True,
    ) -> Promise[Result]:
        """
        Run the scheduler to evaluate an Expression `expr`.

        This is a private method that returns the original workflow Promise.
        """
        # Set scheduler and start executors.

        try:
            set_current_scheduler(self)
            self.clear()

            for executor in self.executors.values():
                executor.start()
            self._dryrun = dryrun
            self._use_cache = cache
            self.traceback = None

            self._validate_tasks()

            # Start event loop to evaluate expression.
            start_time = time.time()
            self.thread_id = threading.get_ident()
            result = self.evaluate(expr, parent_job=parent_job)
            self._process_events(result)

            # Log execution duration.
            duration = time.time() - start_time
            self.log(f"Execution duration: {duration:.2f} seconds")
        except Exception:
            raise
        finally:
            # Stop executors and unset scheduler.
            for executor in self.executors.values():
                executor.stop()

            self._current_execution = None
            set_current_scheduler(None)

        return result

    @overload
    def run(
        self,
        expr: Expression[Result],
        exec_argv: Optional[List[str]] = None,
        dryrun: bool = False,
        cache: bool = True,
        tags: Iterable[Tuple[str, Any]] = (),
    ) -> Result:
        pass

    @overload
    def run(
        self,
        expr: Result,
        exec_argv: Optional[List[str]] = None,
        dryrun: bool = False,
        cache: bool = True,
        tags: Iterable[Tuple[str, Any]] = (),
    ) -> Result:
        pass

    def run(
        self,
        expr: Union[Expression[Result], Result],
        exec_argv: Optional[List[str]] = None,
        dryrun: bool = False,
        cache: bool = True,
        tags: Iterable[Tuple[str, Any]] = (),
    ) -> Result:
        """
        Run the scheduler to evaluate an Expression `expr`.

        This is the primary user entry point to the scheduler. The scheduler will register
        itself as this thread's running scheduler for the duration of the execution.
        This function blocks, running the event loop until the result is ready and can be returned.

        The scheduler can only run one execution at a time.
        """
        if needs_root_task(self.task_registry, expr):
            # Ensure we always have one root-level job encompassing the whole execution.
            expr = root_task(quote(expr))

        # Initialize execution.
        parent_job: Optional[Job]
        if exec_argv is None:
            exec_argv = ["scheduler.run", trim_string(repr(expr))]
        self._current_execution = Execution(self.backend.record_execution(exec_argv))
        self.backend.record_tags(
            TagEntity.Execution, self._current_execution.id, chain(self._exec_tags, tags)
        )

        self.log(
            "Start Execution {exec_id}:  redun {argv}".format(
                exec_id=self._current_execution.id,
                argv=" ".join(map(shlex.quote, exec_argv[1:])),
            )
        )

        result: Promise[Result] = self._run(
            cast(TaskExpression[Result], expr), dryrun=dryrun, cache=cache
        )

        # Return or raise result depending on whether it succeeded.
        if result.is_fulfilled:
            return result.value

        elif result.is_rejected:
            if self.traceback:
                # Log traceback.
                self.log("*** Execution failed. Traceback (most recent task last):")
                for line in self.traceback.format():
                    self.log(line.rstrip("\n"))

            raise result.error

        elif result.is_pending and self._dryrun:
            self.log("Dryrun: Additional jobs would run.")
            raise DryRunResult()

        else:
            raise AssertionError("Unexpected state")

    def extend_run(
        self,
        expr: Union[Expression[Result], Result],
        parent_job_id: str,
        dryrun: bool = False,
        cache: bool = True,
    ) -> dict:
        """
        Extend an existing scheduler execution (run) to evaluate a Task or Expression.

        This is an alternative to the `run` method, and acts as a primary user entry point.
        """
        if needs_root_task(self.task_registry, expr):
            # Ensure we always have one root-level job encompassing the whole execution.
            expr = root_task(quote(expr))

        # Extend an existing execution.
        parent_job_details = self.backend.get_job(parent_job_id)
        if not parent_job_details:
            raise ValueError(f"Unknown parent_job_id: {parent_job_id}")
        self._current_execution = Execution(parent_job_details["execution_id"])
        # Create dummy parent_job to mimic job in the parent scheduler:
        # - We need to set its id to match the given parent_job_id.
        # - We set the same execution id as the original parent_job.
        # - Any expression will do, so we use a minimal root task.
        # - We also mimic the task and eval_args being set like they are in _exec_job().
        parent_job = Job(
            root_task,
            root_task(quote(None)),  # type: ignore
            id=parent_job_id,
            execution=self._current_execution,
        )
        parent_job.eval_args = ((quote(None),), {})

        result: Promise[Result] = self._run(
            cast(TaskExpression[Result], expr), parent_job=parent_job, dryrun=dryrun, cache=cache
        )

        # There will always be one new job id for this subexecution.
        [job] = parent_job.child_jobs

        # Return the result along with the job id to allow caller to stitch the job tree.
        if result.is_fulfilled:
            return {
                "result": result.value,
                "job_id": job.id,
                "call_hash": job.call_hash,
            }
        elif result.is_rejected:
            return {
                "error": result.error,
                "job_id": job.id,
                "call_hash": job.call_hash,
            }
        elif result.is_pending and self._dryrun:
            # Return an incomplete run due to dryrun being active.
            return {"dryrun": True, "job_id": job.id, "call_hash": job.call_hash}
        else:
            raise AssertionError("Unexpected state")

    def _process_events(self, workflow_promise: Promise) -> None:
        """
        Main scheduler event loop for evaluating the current expression. Loop over events,
        blocking until the provided `workflow_promise` is resolved.
        """
        self.workflow_promise = workflow_promise

        while self.workflow_promise.is_pending:
            if self._dryrun and self.events_queue.empty():
                # We have exhausted the completed part of the workflow.
                break

            try:
                event_func = self.events_queue.get(timeout=self.job_status_interval)
                event_func()
            except KeyboardInterrupt:
                self.log("Shutting down... Ctrl+C again to force shutdown.")
                sys.exit(1)
            except queue.Empty:
                if not is_debugger_active():
                    # Print job statuses periodically.
                    # For convenience, do not print statuses if the debugger is
                    # active since they interfere with the debugger REPL.
                    self.log_job_statuses()

        # Print final job statuses.
        self.log_job_statuses()

    def get_job_status_report(self) -> List[str]:
        # Gather job status information.
        status_counts: Dict[str, Dict[str, int]] = defaultdict(lambda: defaultdict(int))
        for job in self._jobs:
            status_counts[job.task.fullname][job.status] += 1
            status_counts[job.task.fullname]["TOTAL"] += 1
        for task_name, task_status_counts in self._finalized_jobs.items():
            for status, count in task_status_counts.items():
                status_counts[task_name][status] += count
                status_counts[task_name]["TOTAL"] += count

        return list(format_job_statuses(status_counts))

    def log_job_statuses(self) -> None:
        """
        Display Job statuses.
        """
        self.log()
        for report_line in self.get_job_status_report():
            self.log(report_line)
        self.log()

    def evaluate(self, expr: AnyExpression, parent_job: Optional[Job] = None) -> Promise:
        """
        Begin an asynchronous evaluation of an expression (concrete value or Expression). Assumes
        this scheduler is currently running.

        This method is not a typical entry point for users, however, it is often used by
        `SchedulerTask`s, to trigger further computations.

        Returns a Promise that will resolve when the evaluation is complete.
        """

        assert self.is_running, "Scheduler is not currently running."

        def eval_term(value):
            # Evaluate one term of an expression.
            if isinstance(value, ValueExpression):
                return value.value
            elif isinstance(value, ApplyExpression):
                return self._evaluate_apply(value, parent_job=parent_job)
            else:
                return value

        def resolve_term(value):
            # Replace promises with their resolved values.
            if isinstance(value, Promise):
                return value.value
            else:
                return value

        pending_expr = map_nested_value(eval_term, expr)
        pending_terms = [
            arg for arg in iter_nested_value(pending_expr) if isinstance(arg, Promise)
        ]
        return Promise.all(pending_terms).then(
            lambda _: map_nested_value(resolve_term, pending_expr)
        )

    def _evaluate_apply(self, expr: ApplyExpression, parent_job: Optional[Job] = None) -> Promise:
        """
        Begin an evaluation of an ApplyExpression.

        This method is responsible for ensuring that each unique expression object is evaluated
        exactly once.

        Returns a Promise that will resolve/reject when the evaluation is complete.
        """

        # Detect expression cycles. Specifically, we look for previous expressions
        # being evaluated within the same parent_job. If we find one, we can
        # reuse its Promise and then copy over the evaluation bookkeeping
        # (call_hash, _upstreams) after its completion.
        # Note: We use `expr.get_hash()` to compensate for an Expression serialization
        # issue discussed in https://github.com/insitro/redun-private/pull/191
        promise_expr = self._pending_expr[parent_job].get(expr.get_hash())
        if promise_expr:
            pending_promise, expr2 = promise_expr

            def callback(result):
                # Copy the evaluation bookkeeping from the completed expression `expr2`
                # to our detected duplicate expression `expr`.
                if isinstance(expr2, TaskExpression):
                    expr.call_hash = expr2.call_hash
                elif isinstance(expr2, SimpleExpression):
                    expr._upstreams = expr2._upstreams
                else:
                    raise AssertionError(f"Unexpected expression: {expr2}")
                return result

            return pending_promise.then(callback)

        # Implementation note: we have to store this promise on the expression provided.
        # This happens at the end of the function, so do not return early!
        promise: Promise

        if isinstance(expr, SchedulerExpression):
            task = self.task_registry.get(expr.task_name)
            if not task:
                promise = Promise(
                    lambda resolve, reject: reject(
                        NotImplementedError(
                            "Scheduler task '{}' does not exist".format(expr.task_name)
                        )
                    )
                )
            else:
                # Scheduler tasks are executed with access to scheduler, parent_job, and
                # raw args (i.e. possibly unevaluated).
                promise = task.func(self, parent_job, expr, *expr.args, **expr.kwargs)

        elif isinstance(expr, TaskExpression):
            # Evaluate task_name to specific task.
            # TaskRegistry is like an environment.
            task = self.task_registry.get(expr.task_name)
            if not task:
                promise = Promise(
                    lambda resolve, reject: reject(
                        f"Task not found in registry: {expr.task_name}.  This can occur if the "
                        "script that defines a user task wasn't loaded or, in the case of redun "
                        "tasks, if an executor is loading a different version of redun that "
                        "fails to define the task."
                    )
                )
            else:
                # TaskExpressions need to be executed in a new Job.
                job = Job(task, expr, parent_job=parent_job, execution=self._current_execution)
                self._jobs.add(job)

                # Make default arguments explicit in case they need to be evaluated.
                expr_args = set_arg_defaults(task, expr.args, expr.kwargs)

                # Evaluate args then execute job.
                args_promise = self.evaluate(expr_args, parent_job=parent_job)
                promise = args_promise.then(lambda eval_args: self._exec_job(job, eval_args))

        elif isinstance(expr, SimpleExpression):
            # Simple Expressions can be executed synchronously.
            func = get_lazy_operation(expr.func_name)
            if not func:
                promise = Promise(
                    lambda resolve, reject: reject(
                        NotImplementedError(
                            "Expression function '{}' does not exist".format(expr.func_name)
                        )
                    )
                )
            else:
                # Evaluate args, apply func, evaluate result.
                args_promise = self.evaluate((expr.args, expr.kwargs), parent_job=parent_job)
                promise = args_promise.then(
                    lambda eval_args: self.evaluate(
                        cast(Callable, func)(*eval_args[0], **eval_args[1]), parent_job=parent_job
                    )
                )

        else:
            raise NotImplementedError("Unknown expression: {}".format(expr))

        # Recording pending expressions.
        self._pending_expr[parent_job][expr.get_hash()] = (promise, expr)
        return promise

    def _is_job_within_limits(self, job_limits: dict) -> bool:
        """
        Helper to determine if a job can be executed under allowed/used limits.
        """
        # Job limits not specified in the config will default to an available limit of 1.
        return all(
            self.limits.get(limit_name, 1) - self.limits_used[limit_name] - count >= 0
            for limit_name, count in job_limits.items()
        )

    def _consume_resources(self, job_limits: Dict[str, int]) -> None:
        """
        Increments the resource limits used with the supplied job limits.
        """
        for limit_name, count in job_limits.items():
            self.limits_used[limit_name] += count

    def _release_resources(self, job_limits: Dict[str, int]) -> None:
        """
        Decrements the resource limits used with the supplied job limits.
        """
        for limit_name, count in job_limits.items():
            self.limits_used[limit_name] -= count

    def _add_job_pending_limits(self, job: Job, eval_args: Tuple[Tuple, dict]) -> None:
        """
        Adds a job to the queue of jobs waiting to run once resources are available.
        """
        self._jobs_pending_limits.append((job, eval_args))

    def _check_jobs_pending_limits(self) -> None:
        """
        Checks whether Jobs pending due to limits can now satisfy limits and run.
        """

        def _add_limits(limits1, limits2):
            return {
                limit_name: limits1[limit_name] + limits2[limit_name]
                for limit_name in set(limits1) | set(limits2)
            }

        ready_jobs: List[Tuple[Job, Tuple[tuple, dict]]] = []
        not_ready_jobs: List[Tuple[Job, Tuple[tuple, dict]]] = []
        ready_job_limits: Dict[str, int] = defaultdict(int)

        # Resource limits that will be consumed by ready jobs.
        for job, eval_args in self._jobs_pending_limits:
            job_limits = job.get_limits()
            proposed_limits = _add_limits(job_limits, ready_job_limits)
            if self._is_job_within_limits(proposed_limits):
                ready_jobs.append((job, eval_args))
                ready_job_limits = proposed_limits
            else:
                not_ready_jobs.append((job, eval_args))

        self._jobs_pending_limits = not_ready_jobs

        # We're just nominating these jobs to start again, but their resource needs will be
        # checked again as part of restarting them. It's possible they'll get re-queued again.
        for job, eval_args in ready_jobs:
            self._exec_job(job, eval_args)

    def _check_pending_job(self, job: Job) -> Optional[Job]:
        """
        Checks whether an equivalent Job is already running.

        This check is necessary to avoid double evaluating an expression that is
        not in the cache yet since it is pending.

        This is a form of common subexpression elimination.
        https://en.wikipedia.org/wiki/Common_subexpression_elimination

        If there is another job we can rely on, return it. Otherwise, return `None`.
        """

        if (
            job.get_option("cache_scope", CacheScope.BACKEND, as_type=CacheScope)
            == CacheScope.NONE
        ):
            return None

        # Check the task has been configured to restrict CSE. This isn't typical, but it
        # is allowed.
        allowed_cache_results = job.get_option("allowed_cache_results", None)
        if allowed_cache_results is not None:
            assert isinstance(allowed_cache_results, set)
            if CacheResult.CSE not in allowed_cache_results:
                return None

        pending_job = self._pending_jobs.get(job.eval_hash)
        if pending_job:
            job.collapse(pending_job)
            return pending_job

        return None

    def _exec_job(self, job: Job, eval_args: Tuple[Tuple, dict]) -> Promise:
        """
        Execute a job that is ready using the fully evaluated arguments.
        """
        # Delay the execution to a new event so we don't interrupt the current event.
        self.events_queue.put(lambda: self._exec_job_main_thread(job, eval_args))
        return job.result_promise

    def _exec_job_main_thread(self, job: Job, eval_args: Tuple[Tuple, dict]) -> None:
        """
        Execute a job that is ready using the fully evaluated arguments.

        This function runs on the main scheduler thread.
        """

        # Downgrade the cache option, if needed.
        if not self._use_cache:
            job.task_options["cache_scope"] = CacheScope.NONE

        # Ensure we are on main scheduler thread.
        assert self.thread_id == threading.get_ident()
        assert job.task

        assert not job.was_cached, f"Job {job} was marked as cached, but we're evaluating it?"

        # Set eval_args on job.
        job.eval_args = eval_args

        # Preprocess arguments before sending them to task function.
        args, kwargs = job.eval_args
        args, kwargs = job.args = self._preprocess_args(job, args, kwargs)

        # Check cache using eval_hash as key.
        job.eval_hash, job.args_hash = self.get_eval_hash(job.task, args, kwargs)
        assert job.eval_hash
        assert job.args_hash

        # Check whether a job of the same eval_hash is already running.
        if self._check_pending_job(job) is not None:
            # If we're trying to execute a job that is being re-triggered because we hit
            # resource limits on the first try, it's important we don't hit this exit,
            # because then the job will have been dropped.

            # There's no work to do, but be sure we consider it started.
            self.backend.record_job_start(job)

            return

        # Check cache for job.
        result, job.was_cached, call_hash = self._get_cache(job)

        if job.was_cached:
            # Evaluation was cached, so we proceed to done_job/resolve_job.
            self.log(
                "{action} Job {job_id}:  {task_call} (eval_hash={eval_hash}{check_valid}, "
                "call_hash={call_hash})".format(
                    job_id=job.id[:8],
                    action="Cached".ljust(JOB_ACTION_WIDTH),
                    task_call=format_task_call(job.task, args, kwargs),
                    eval_hash=job.eval_hash[:8],
                    call_hash=call_hash[:8] if call_hash else "None",
                    check_valid=", check_valid=shallow" if call_hash else "",
                )
            )

            # Record the call hash, if we recovered one.
            job.call_hash = call_hash

            # There's no work to do, but be sure we consider it started.
            self.backend.record_job_start(job)

            # Trigger downstream steps, just like an executor would, upon completing it.
            # One of the roles of `done_job` is to trigger evaluation on `result`, in case it is
            # an expression. In the case of an ultimate reduction cache hit, we know that's not
            # needed. However, we let this duplicate work happen anyway, because it keeps the
            # program flow the same and is inexpensive.
            if isinstance(result, ErrorValue):
                return self.reject_job(job, result.error, result.traceback)
            else:
                return self.done_job(job, result)

        # Perform rollbacks due to Handles that conflict with past Handles.
        if not self._dryrun:
            self._perform_rollbacks(args, kwargs)

        # Determine executor.
        executor_name = job.get_option("executor") or "default"
        executor = self.executors.get(executor_name)
        if not executor:
            return self.reject_job(
                job, SchedulerError('Unknown executor "{}"'.format(executor_name))
            )

        self.log(
            "{action} Job {job_id}:  {task_call} on {executor}".format(
                job_id=job.id[:8],
                action=("Dryrun" if self._dryrun else "Run").ljust(JOB_ACTION_WIDTH),
                task_call=format_task_call(job.task, args, kwargs),
                executor=executor_name,
            )
        )

        # Stop short of submitting jobs during a _dryrun.
        if self._dryrun:
            return

        # Since the result was not cached, either run it now, or reschedule for whenever
        # the resources are available.
        job_limits = job.get_limits()
        if not self._is_job_within_limits(job_limits):
            self._add_job_pending_limits(job, eval_args)
            return
        self._consume_resources(job_limits)

        # Record that the job is actually starting.
        self.backend.record_job_start(job)

        # Record the job as pending, since we're submitting it.
        # Note that if the CSE is disabled, this job might have the same `eval_hash` as a prior
        # one. We don't care about overwriting, however, since they're all equivalent.
        self._pending_jobs[job.eval_hash] = job

        # Submit job.
        if not job.task.script:
            executor.submit(job)
        else:
            executor.submit_script(job)

    def done_job(self, job: Job, result: Any, job_tags: List[Tuple[str, Any]] = []) -> None:
        """
        Mark a :class:`Job` as successfully done with a `result`.

        A primary Executor lifecycle method, hence is thread safe.
        """
        self.events_queue.put(lambda: self._done_job_main_thread(job, result, job_tags=job_tags))

    def _done_job_main_thread(
        self, job: Job, result: Any, job_tags: List[Tuple[str, Any]] = []
    ) -> None:
        """
        Mark a :class:`Job` as successfully done with a `result`.

        Note that `result` might require additional evaluation.

        This function runs on the main scheduler thread. Use
        :method:`Scheduler.done_job()` if calling from another thread.
        """

        # Ensure we are on main scheduler thread.
        assert self.thread_id == threading.get_ident()

        # Cached jobs won't have used any resources.
        if not job.was_cached:
            self._release_resources(job.get_limits())
            self._check_jobs_pending_limits()

        assert job.task
        assert job.eval_hash
        assert job.eval_args
        assert job.args_hash

        job.job_tags.extend(job_tags)

        # Update cache. Note that cache scope dictates whether we *read* from the cache, but we
        # always *write* to the backend.
        if not job.was_cached:
            assert not self._dryrun
            result = self._postprocess_result(job, result, job.eval_hash)
            self.set_cache(job.eval_hash, job.task.hash, job.args_hash, result)

        # Eval result (child tasks), then resolve job.
        self.evaluate(result, parent_job=job).then(
            lambda result: self._resolve_job(job, result)
        ).catch(lambda error: self.reject_job(job, error))

    def _resolve_job(self, job: Job, result: Any) -> None:
        """
        Resolve a :class:`Job` with a fully evaluated result. Internal method; executors should
        call `done_job` or `reject_job`.

        This occurs after a job is done and the result is fully evaluated.
        """
        self.events_queue.put(lambda: self._resolve_job_main_thread(job, result))

    def _resolve_job_main_thread(self, job: Job, result: Any) -> None:
        """
        Resolve a :class:`Job` with a fully evaluated result.

        This occurs after a job is done and the result is fully evaluated.

        This function runs on the main scheduler thread. Use
        :method:`Scheduler._resolve_job()` if calling from another thread.
        """
        # Ensure we are on main scheduler thread.
        assert self.thread_id == threading.get_ident()

        assert job.task
        assert job.args_hash
        assert job.eval_args

        # If the call hash is already known, we must have retrieved it from the cache. Otherwise,
        # compute the hash and record the call node.
        if job.call_hash:
            assert job.was_cached
        else:
            # Ignore failed child jobs, which have no call_hash.
            child_call_hashes = [
                cast(str, child_job.call_hash)
                for child_job in job.child_jobs
                if child_job.call_hash
            ]

            # Compute final call_hash and record CallNode.
            # This is a no-op if job was cached.
            result_hash = self.backend.record_value(result)
            job.call_hash = self.backend.record_call_node(
                task_name=job.task.fullname,
                task_hash=job.task.hash,
                args_hash=job.args_hash,
                expr_args=(job.expr.args, job.expr.kwargs),
                eval_args=job.eval_args,
                result_hash=result_hash,
                child_call_hashes=child_call_hashes,
            )

        self._record_job_tags(job)

        self.backend.record_job_end(job)
        job.resolve(result)
        self._finalize_job(job)

    def _finalize_job(self, job: Job) -> None:
        """
        Clean up finalized job.
        """
        self._jobs.remove(job)
        self._finalized_jobs[job.task.fullname][job.status] += 1

        # By poping the job, we free up all expressions that we created within
        # the job. They no longer can be part of any more cycles.
        self._pending_expr.pop(job, None)

        # Once a job has been recorded to the cache, we don't need to keep around
        # the job, since if we see it again, we'll simply download the result.
        self._pending_jobs.pop(job.eval_hash, None)

    def _record_job_tags(self, job: Job) -> None:
        """
        Record tags acquired during Job.
        """
        assert job.task
        assert job.execution

        # Record value tags.
        for value_hash, tags in job.value_tags:
            self.backend.record_tags(entity_type=TagEntity.Value, entity_id=value_hash, tags=tags)

        # Record Job tags.
        job_tags = job.get_option("tags", []) + job.job_tags
        if job_tags:
            self.backend.record_tags(entity_type=TagEntity.Job, entity_id=job.id, tags=job_tags)

        # Record Execution tags.
        if job.execution_tags:
            self.backend.record_tags(
                entity_type=TagEntity.Execution,
                entity_id=job.execution.id,
                tags=job.execution_tags,
            )

        # Record Task tags.
        task_tags = job.task.get_task_option("tags")
        if task_tags:
            self.backend.record_tags(
                entity_type=TagEntity.Task, entity_id=job.task.hash, tags=task_tags
            )

    def reject_job(
        self,
        job: Optional[Job],
        error: Any,
        error_traceback: Optional[Traceback] = None,
        job_tags: List[Tuple[str, Any]] = [],
    ) -> None:
        """
        Reject a :class:`Job` that has failed with an `error`.

        A primary Executor lifecycle method, hence is thread safe.
        """
        self.events_queue.put(
            lambda: self._reject_job_main_thread(
                job, error, error_traceback=error_traceback, job_tags=job_tags
            )
        )

    def _reject_job_main_thread(
        self,
        job: Optional[Job],
        error: Any,
        error_traceback: Optional[Traceback] = None,
        job_tags: List[Tuple[str, Any]] = [],
    ) -> None:
        """
        Reject a :class:`Job` that has failed with an `error`.

        This function runs on the main scheduler thread. Use
        :method:`Scheduler.reject_job()` if calling from another thread.
        """
        # Ensure we are on main scheduler thread.
        assert self.thread_id == threading.get_ident()

        # Attach the traceback to the error so the redun traceback is always available on the error
        # itself. In some cases (like AWS batch), the error will be constructed by unpickling a
        # pickled Excetption(__traceback__ is not included when pickling). By attaching the
        # traceback here, we make sure the traceback is available for downstream handlers.
        error.redun_traceback = error_traceback

        if job:
            assert job.task
            assert job.args_hash
            assert job.eval_args

            job.job_tags.extend(job_tags)

            self.log(
                "*** {action} Job {job_id}:  {task_call}:".format(
                    job_id=job.id[:8],
                    action="Reject".ljust(JOB_ACTION_WIDTH),
                    task_call=format_task_call(job.task, job.eval_args[0], job.eval_args[1]),
                )
            )

            # Cached jobs won't have used any resources.
            if not job.was_cached:
                self._release_resources(job.get_limits())
                self._check_jobs_pending_limits()

            if self.use_task_traceback:
                self._set_task_traceback(job, error, error_traceback=error_traceback)

            child_call_hashes = [
                child_job.call_hash for child_job in job.child_jobs if child_job.call_hash
            ]

            # Compute final call_hash and record CallNode.
            error_value = ErrorValue(error, error_traceback or Traceback.from_error(error))
            try:
                error_hash = self.backend.record_value(error_value)
            except (TypeError, AttributeError):
                # Some errors cannot be serialized so record them as generic Exceptions.
                error2 = Exception(repr(error))
                error_value = ErrorValue(error2, error_traceback or Traceback.from_error(error2))
                error_hash = self.backend.record_value(error_value)
            job.call_hash = self.backend.record_call_node(
                task_name=job.task.fullname,
                task_hash=job.task.hash,
                args_hash=job.args_hash,
                expr_args=(job.expr.args, job.expr.kwargs),
                eval_args=job.eval_args,
                result_hash=error_hash,
                child_call_hashes=child_call_hashes,
            )
            self._record_job_tags(job)
            self.backend.record_job_end(job, status="FAILED")
            job.reject(error)
            self._finalize_job(job)

        else:
            # Error is not job specific. Often this could be a redun-internal error.
            # Stop whole workflow.
            self.log("*** Workflow error")
            if not error_traceback:
                # Since this is an unusual workflow error, lets see all the frames.
                error_traceback = Traceback.from_error(error, trim_frames=False)
            self.traceback = error_traceback
            assert self.workflow_promise
            self.workflow_promise.do_reject(error)

    def _set_task_traceback(
        self, job: Job, error: Any, error_traceback: Optional[Traceback] = None
    ) -> None:
        """
        Set traceback to be :class:`redun.task.Task` based.
        """
        if self.traceback:
            # The traceback has already been set.
            return

        # Get job stack.
        job_stack: List[Job] = []
        ptr: Optional[Job] = job
        while ptr:
            job_stack.append(ptr)
            ptr = ptr.parent_job
        job_stack.reverse()

        # Get task-based traceback.
        task_frames: List[FrameSummary] = []
        for frame_job in job_stack:
            assert frame_job.task
            assert frame_job.eval_args
            args, kwargs = frame_job.eval_args

            sig = inspect.signature(frame_job.task.func)
            task_frames.append(
                Frame(
                    frame_job.task.func.__code__.co_filename,
                    frame_job.task.func.__code__.co_firstlineno,
                    frame_job.task.func.__name__,
                    job=frame_job,
                    locals={
                        **dict(zip(sig.parameters, args)),
                        **kwargs,
                    },
                )
            )

        # Graft user function frames onto task traceback.
        if not error_traceback:
            error_traceback = Traceback.from_error(error)
        task_frames.extend(error_traceback.frames)

        # Set this execution's traceback.
        self.traceback = Traceback(error, task_frames, logs=error_traceback.logs)

    def get_eval_hash(self, task: Task, args: tuple, kwargs: dict) -> Tuple[str, str]:
        """
        Compute eval_hash and args_hash for a task call, filtering out config args before hashing.
        """
        # Filter out config args from args and kwargs.
        sig = task.signature
        config_args: List = task.get_task_option("config_args", [])

        # Determine the variadic parameter if it exists.
        var_param_name: Optional[str] = None
        for param in sig.parameters.values():
            if param.kind == inspect.Parameter.VAR_POSITIONAL:
                var_param_name = param.name
                break

        # Filter args to remove config_args.
        args2 = [
            arg_value
            for arg_name, arg_value in zip(sig.parameters, args)
            if arg_name not in config_args
        ]

        # Additional arguments are assumed to be variadic arguments.
        args2.extend(
            arg_value
            for arg_value in args[len(sig.parameters) :]
            if var_param_name not in config_args
        )

        # Filter kwargs.
        kwargs2 = {
            arg_name: arg_value
            for arg_name, arg_value in kwargs.items()
            if arg_name not in config_args
        }

        return hash_eval(self.type_registry, task.hash, args2, kwargs2)

    def _get_cache(self, job: Job) -> Tuple[Any, bool, Optional[str]]:
        """
        Attempt to find a cached value for an evaluation (`eval_hash`).

        Returns
        -------
        (result, is_cached, call_hash): Tuple[Any, bool, Optional[str]]
           is_cached is True if `result` was in the cache.
           Note that the result may be an `ErrorValue`, which the caller must unwrap.
        """
        check_valid = job.get_option("check_valid", CacheCheckValid.FULL, as_type=CacheCheckValid)

        # HACK: This is a workaround introduced by DE-4761. See the ticket for
        # notes on how script_task could be redesigned to remove the need for
        # this special-casing.
        is_script_task = job.task.fullname == "redun.script_task"

        # Determine whether we should use cache.
        cache_scope = (
            CacheScope.CSE
            if is_script_task
            else job.get_option("cache_scope", CacheScope.BACKEND, as_type=CacheScope)
        )

        allowed_cache_results = job.get_option("allowed_cache_results", None)
        assert allowed_cache_results is None or isinstance(allowed_cache_results, set), (
            "Should have gotten a set for option `allowed_cache_results`, got"
            f" {allowed_cache_results} instead."
        )

        # Check CSE and cache.
        assert job.task
        assert job.args_hash
        assert job.eval_hash
        assert job.execution
        result, call_hash, cache_type = self.backend.check_cache(
            job.task.hash,
            job.args_hash,
            job.eval_hash,
            job.execution.id,
            self.task_registry.task_hashes,
            cache_scope,
            check_valid,
            allowed_cache_results,
        )

        if cache_type == CacheResult.CSE:
            # If this is a CSE hit, we can use the result immediately. The result may be
            # an error wrapped as a `ErrorValue`, but we can still use it.
            return result, True, call_hash
        elif isinstance(result, ErrorValue):
            # Errors can't be used from the backend cache.
            return None, False, None
        elif cache_type == CacheResult.MISS:
            if self._dryrun:
                # In dryrun mode, log reason for cache miss.
                self._log_cache_miss(job, job.task.hash, job.args_hash)
            return None, False, None

        elif self._is_valid_value(result):
            # Result must still be valid to use.
            return result, True, call_hash

        else:
            self.log(
                "{action} Job {job}:  Cached result is no longer valid "
                "(result={result}, eval_hash={eval_hash}).".format(
                    action="Miss".ljust(JOB_ACTION_WIDTH),
                    job=job.id[:8],
                    result=trim_string(repr(result)),
                    eval_hash=job.eval_hash[:8],
                )
            )
            return None, False, None

    def _log_cache_miss(self, job: Job, task_hash: str, args_hash: str) -> None:
        """
        Log the reason for a cache miss.
        """
        task = self.task_registry.get(hash=task_hash)
        reason = self.backend.explain_cache_miss(task, args_hash)

        if not reason:
            # No reason could be determined.
            self.log(
                "{action} Job {job}:  Cannot determine reason for cache miss of "
                "'{task_name}()' (task_hash={task_hash}, args_hash={args_hash}).".format(
                    action="Miss".ljust(JOB_ACTION_WIDTH),
                    job=job.id[:8],
                    task_name=task.fullname,
                    task_hash=task_hash[:8],
                    args_hash=args_hash[:8],
                )
            )

        elif reason["reason"] == "new_task":
            self.log(
                "{action} Job {job}:  New task '{task_name}()' with previous arguments "
                "(task_hash={old_hash} --> {new_hash}, args_hash={args_hash}).".format(
                    action="Miss".ljust(JOB_ACTION_WIDTH),
                    job=job.id[:8],
                    task_name=task.fullname,
                    old_hash=reason["call_task_hash"][:8],
                    new_hash=task_hash[:8],
                    args_hash=args_hash[:8],
                )
            )

        elif reason["reason"] == "new_args":
            self.log(
                "{action} Job {job}:  Existing task '{task_name}()' is called with new arguments "
                "(task_hash={task_hash}, args_hash={old_hash} --> {new_hash}).".format(
                    action="Miss".ljust(JOB_ACTION_WIDTH),
                    job=job.id[:8],
                    task_name=task.fullname,
                    task_hash=task_hash[:8],
                    old_hash=reason["call_args_hash"][:8],
                    new_hash=args_hash[:8],
                )
            )

        elif reason["reason"] == "new_call":
            self.log(
                "{action} Job {job}:  New task '{task_name}()' is called with new arguments "
                "(task_hash={task_hash}, args_hash={args_hash}).".format(
                    action="Miss".ljust(JOB_ACTION_WIDTH),
                    job=job.id[:8],
                    task_name=task.fullname,
                    task_hash=task_hash[:8],
                    args_hash=args_hash[:8],
                )
            )

        else:
            raise NotImplementedError(reason)

    def set_cache(self, eval_hash: str, task_hash: str, args_hash: str, value: Any) -> None:
        """
        Set the cache for an evaluation (`eval_hash`) with result `value`.

        This method should only be used by the scheduler or `SchedulerTask`s
        """
        self.backend.set_eval_cache(eval_hash, task_hash, args_hash, value, value_hash=None)

    def _is_valid_value(self, value: Any) -> bool:
        """
        Returns True if the value is valid.

        Valid Nested Values must have all of their subvalues be valid.
        """
        return self.type_registry.is_valid_nested(value)

    def _perform_rollbacks(self, args: Tuple, kwargs: dict) -> None:
        """
        Perform any Handle rollbacks needed for the given Task arguments.
        """
        for value in iter_nested_value((args, kwargs)):
            if isinstance(value, Handle):
                self.backend.rollback_handle(value)

    def _preprocess_args(self, job: Job, args: Tuple, kwargs: dict) -> Any:
        """
        Preprocess arguments for a Task before execution.
        """

        def preprocess_value(value):
            if isinstance(value, Handle):
                if job.parent_job:
                    job.parent_job.handle_forks[value.get_hash()] += 1
                    call_order = job.parent_job.handle_forks[value.get_hash()]
                else:
                    call_order = 0
                preprocess_args = {"call_order": call_order}
            else:
                preprocess_args = {}

            value2 = self.type_registry.preprocess(value, preprocess_args)

            # Handles need extra recording for preprocessing.
            if isinstance(value, Handle):
                assert value2 != value
                self.backend.advance_handle([value], value2)

            return value2

        return map_nested_value(preprocess_value, (args, kwargs))

    def _postprocess_result(self, job: Job, result: Any, pre_call_hash: str) -> Any:
        """
        Postprocess a result from a Task before caching.
        """
        postprocess_args = {
            "pre_call_hash": pre_call_hash,
        }

        def postprocess_value(value):
            value2 = self.type_registry.postprocess(value, postprocess_args)

            if isinstance(value, Handle) and not self._dryrun:
                # Handles accumulate state change from jobs that emit them.
                assert value2 != value
                self.backend.advance_handle([value], value2)

            return value2

        return map_nested_value(postprocess_value, result)


@scheduler_task(namespace="redun")
def merge_handles(
    scheduler: Scheduler, parent_job: Job, sexpr: SchedulerExpression, handles_expr: List[Handle]
) -> Promise:
    """
    Merge multiple handles into one.
    """

    def then(handles):
        assert handles
        assert len({handle.__handle__.name for handle in handles}) == 1

        final_handle = handles[0]
        other_handles = handles[1:]

        scheduler.backend.advance_handle(other_handles, final_handle)
        return final_handle

    return scheduler.evaluate(handles_expr, parent_job=parent_job).then(then)


@scheduler_task(namespace="redun")
def cond(
    scheduler: Scheduler,
    parent_job: Job,
    sexpr: SchedulerExpression,
    cond_expr: Any,
    then_expr: Any,
    *rest: Any,
) -> Promise:
    """
    Conditionally execute expressions, i.e. a lazy if-statement.

    .. code-block:: python

        @task()
        def is_even(x):
            return x % 2 == 0

        @task()
        def task1():
            # ...

        @task()
        def task2():
            # ...

        @task()
        def main(x: int):
            return cond(
                is_even(x),
                task1(),
                task2()
            )
    """
    exprs = (cond_expr, then_expr) + rest

    def then(args):
        i, cond_value = args

        if cond_value:
            # Return 'then' clause.
            return scheduler.evaluate(exprs[i + 1], parent_job=parent_job)

        elif len(exprs) - i == 3:
            # No more expresses, so return 'otherwise' clause.
            return scheduler.evaluate(exprs[i + 2], parent_job=parent_job)

        else:
            # Recurse to next conditional clause.
            return scheduler.evaluate((i + 2, exprs[i + 2]), parent_job=parent_job).then(then)

    # Evaluate conditional clause.
    return scheduler.evaluate((0, cond_expr), parent_job=parent_job).then(then)


@task(namespace="redun", version="1")
def throw(error: Exception) -> None:
    """
    Raises an exception.

    This is task is useful in cases where raising needs to be done lazily.
    """
    raise error


@scheduler_task(namespace="redun")
def catch(
    scheduler: Scheduler, parent_job: Job, sexpr: SchedulerExpression, expr: Any, *catch_args: Any
) -> Promise:
    """
    Catch exceptions `error` of class `error_class` from `expr` and evaluate `recover(error)`.

    The following example catches the `ZeroDivisionError` raised by `divider(0)`
    and returns 0.0 instead.

    .. code-block:: python

        @task()
        def divider(denom):
            return 1.0 / denom

        @task()
        def recover(error):
            return 0.0

        @task()
        def main():
            return catch(
                divider(0),
                ZeroDivisionError,
                recover
            )

    This is equivalent to the regular python code:

    .. code-block:: python

        def divider(denom):
            return 1.0 / denom

        def main():
            try:
                return denom(0)
            except ZeroDivisionError as error:
                return 0.0

    `catch()` can also handle multiple Exception classes as well as multiple
    recover expressions.

    .. code-block:: python

        @task()
        def main():
            return catch(
                task1(),
                (ZeroDivisionError, ValueError),
                recover1,
                KeyError,
                recover2,
            )

    which is equivalent to the regular python code:

    .. code-block:: python

        try:
            return task1()
        except (ZeroDivisionError, ValueError) as error:
            return recover1(error)
        except KeyError as error:
            return recover2(error)

    Implementation/behavior note: Usually, we don't cache errors, since this is rarely productive.
    Also recall that we typically cache each round of evaluation separately, allowing
    the scheduler to retrace the call graph and detect if any subtasks have changed
    (i.e. hash change) or if any Values in the subworkflow are now invalid (e.g. File hash has
    changed). To correctly cache a caught expression, we need to cache both the fact that
    `expr` resolved to an `Exception`, then that the `recover_expr` produced a non-error result.
    However, we only want to cache the exception if it is successfully handled, so if the
    recovery re-raises the exception or issues another error, the error should not be cached.

    In order to create this behavior, we have to implement custom caching behavior that delays
    caching the error until we see that it has been successfully handled.


    Parameters
    ----------
    expr: Expression
        Main expression to evaluate.
    error_recover_pairs:
        A list of alternating error_class and recover Tasks.
    """
    errors: List[Union[Type[Exception], Tuple[Type[Exception]]]] = list(catch_args[::2])
    recovers: List[Callable] = list(catch_args[1::2])
    error_recover_pairs = list(zip(errors, recovers))
    eval_hash: str
    args_hash: str
    recover_expr: Any = None

    def on_success(result: Result) -> Result:
        # Cache `expr` if catch is ultimately successful.
        scheduler.backend.record_value(catch)
        scheduler.set_cache(eval_hash, catch.hash, args_hash, expr)
        return result

    def on_recover(result: Result) -> Result:
        # Cache `recover_expr` if exception was thrown.
        scheduler.backend.record_value(catch)
        scheduler.set_cache(eval_hash, catch.hash, args_hash, recover_expr)
        return result

    def promise_catch(error: Any) -> Promise:
        nonlocal recover_expr

        for error_class, recover in error_recover_pairs:
            if isinstance(error, error_class):
                # Record dataflow for caught error:
                #   expr --> error --> recover(error) --> sexpr
                error_expr = derive_expression(expr, error)
                recover_expr = recover(error_expr)
                derive_expression(recover_expr, sexpr)

                # Error matches the class, evaluate the recover task.
                return scheduler.evaluate(recover_expr, parent_job=parent_job).then(on_recover)

        # Error did not match any classes, reraise the error.
        raise error

    # Check cache.
    catch_args = (expr,) + catch_args
    eval_hash, args_hash = scheduler.get_eval_hash(catch, catch_args, {})
    # We haven't set an option on the catch task, so we just have to look for one at runtime.
    cache_scope = CacheScope(sexpr.task_expr_options.get("cache_scope", CacheScope.BACKEND))
    if scheduler._use_cache and cache_scope != CacheScope.NONE:
        assert parent_job.execution
        cached_expr, call_hash, cache_type = scheduler.backend.check_cache(
            task_hash=catch.hash,
            args_hash=args_hash,
            eval_hash=eval_hash,
            execution_id=parent_job.execution.id,
            cache_scope=cache_scope,
            check_valid=CacheCheckValid.FULL,
            scheduler_task_hashes=scheduler.task_registry.task_hashes,
            allowed_cache_results={CacheResult.SINGLE},
        )
        if cache_type != CacheResult.MISS:
            return scheduler.evaluate(cached_expr, parent_job=parent_job).catch(promise_catch)

    return scheduler.evaluate(expr, parent_job=parent_job).then(on_success, promise_catch)


@scheduler_task(namespace="redun", version="1")
def catch_all(
    scheduler: Scheduler,
    parent_job: Job,
    sexpr: SchedulerExpression,
    exprs: T,
    error_class: Union[None, Exception, Tuple[Exception, ...]] = None,
    recover: Optional[Task[Callable[..., S]]] = None,
) -> Promise[Union[T, S]]:
    """
    Catch all exceptions that occur in the nested value `exprs`.

    This task works similar to `catch`, except it takes multiple
    expressions within a nested value (e.g. a list, set, dict, etc).
    Any exceptions raised by the expressions are caught and processed only after
    all expressions succeed or fail. For example, this can be useful in large
    fanouts where the user wants to avoid one small error stopping the whole
    workflow immediately. Consider the following code:

    .. code-block:: python

        @task
        def divider(denom):
            return 1.0 / denom

        @task
        def recover(values):
            nerrors = sum(1 for value in values if isinstance(value, Exception))
            raise Exception(f"{nerrors} error(s) occurred.")

        @task
        def main():
            result catch_all([divider(1), divider(0), divider(2)], ZeroDivisionError, recover)

    Each of the expressions in the list `[divider(1), divider(0), divider(2)]`
    will be allowed to finish before the exceptions are processed by `recover`. If
    there are no exceptions, the evaluated list is returned. If there are any
    exceptions, the list is passed to `recover`, where it will contain each
    successful result or Exception. `recover` then has the opportunity to process
    all errors together. In the example above, the total number of errors is
    reraised.

    Parameters
    ----------
    exprs:
        A nested value (list, set, dict, tuple, NamedTuple) of expressions to evaluate.
    error_class: Union[Exception, Tuple[Expression, ...]]
        An Exception or Exceptions to catch.
    recover: Task
        A task to call if any errors occurs. It will be called with the nested
        value containing both successful results and errors.
    """

    def postprocess(pending_terms):
        # Gather all errors.
        errors = [promise.error for promise in pending_terms if promise.is_rejected]

        if errors:
            if not recover:
                # By default, just reraise the first error.
                raise errors[0]
            else:

                def do_recover(error_class_recover):
                    error_class, recover = error_class_recover
                    if all(isinstance(error, error_class) for error in errors):
                        return scheduler.evaluate(
                            recover(map_nested_value(resolve_term, pending_expr)),
                            parent_job=parent_job,
                        )
                    else:
                        # By default, just reraise first non-matching error.
                        raise next(error for error in errors if not isinstance(error, error_class))

                # Evaluate error_class and recover in case they're Expressions.
                return scheduler.evaluate((error_class, recover), parent_job=parent_job).then(
                    do_recover
                )
        else:
            # Return the evaluated nested value.
            return map_nested_value(resolve_term, pending_expr)

    def resolve_term(value):
        # Replace a Promise with its return value or its error.
        if isinstance(value, Promise):
            if value.is_fulfilled:
                return value.value
            else:
                return value.error
        else:
            # Regular values (non-Promises) pass through unchanged.
            return value

    pending_terms = []

    def eval_term(value):
        # Evaluate one term and track its promise in pending_terms.
        promise = scheduler.evaluate(value, parent_job=parent_job)
        pending_terms.append(promise)
        return promise

    pending_expr = map_nested_value(eval_term, exprs)
    return wait_promises(pending_terms).then(postprocess)


@scheduler_task(namespace="redun")
def apply_tags(
    scheduler: Scheduler,
    parent_job: Job,
    sexpr: SchedulerExpression,
    value: Any,
    tags: List[Tuple[str, Any]] = [],
    job_tags: List[Tuple[str, Any]] = [],
    execution_tags: List[Tuple[str, Any]] = [],
) -> Promise:
    """
    Apply tags to a value, job, or execution.

    Returns the original value.

    .. code-block:: python

        @task
        def task1():
            x = 10
            # Apply tags on the value 10.
            return apply_tags(x, [("env": "prod", "project": "acme")])

        @task
        def task2():
            x = 10
            # Apply tags on the current job.
            return apply_tags(x, job_tags=[("env": "prod", "project": "acme")])

        @task
        def task3():
            x = 10
            # Apply tags on the current execution.
            return apply_tags(x, execution_tags=[("env": "prod", "project": "acme")])
    """

    def then(args):
        value, tags, job_tags, execution_tags = args
        if tags:
            value_hash = scheduler.backend.record_value(value)
            parent_job.value_tags.append((value_hash, tags))
        parent_job.job_tags.extend(job_tags)
        parent_job.execution_tags.extend(execution_tags)
        return value

    return scheduler.evaluate((value, tags, job_tags, execution_tags), parent_job=parent_job).then(
        then
    )


@task(
    namespace="redun",
    name="subrun_root_task",
    version="1",
    config_args=["config", "load_modules", "run_config"],
    cache_scope=CacheScope.CSE,
    check_valid=CacheCheckValid.SHALLOW,
)
def _subrun_root_task(
    expr: Any,
    config: Dict[str, Dict],
    load_modules: List[str],
    run_config: Dict[str, Any],
) -> Any:
    """
    Launches a sub-scheduler and runs the provided expression by first "unwrapping" it.
    The evaluated result is returned within a dict alongside other sub-scheduler-related
    state to the caller.

    This task is limited to execution cache scope because it is difficult for the caller to
    be reactive to the details of the subrun. When retrieving a cache value for this task,
    the parent scheduler only sees the final values, not the whole call tree, so it can only
    perform shallow checking. Selecting execution scope means that validity checking is not
    required. We also set shallow checking as a reminder of how it behaves.

    In the case that there is no execution-scope cache hit, then the sub-scheduler has to
    be started. However, the sub-scheduler is free to perform any cache logic it desires. It
    has the complete context to be fully reactive, or not, as configured. So, if there is a
    backend-scope cache hit available, the sub-scheduler may return almost instantly.

    The parent scheduler may or may not have access to the code, which also makes full reactivity
    difficult. It is a mild assumption that the code does not change during an execution.

    There is one particular case we have not implemented, where the parent scheduler could bypass
    the need to start the sub-scheduler. If 1) the user only wants shallow validity checking
    and 2) the parent scheduler has all access all the code (to compute task hashes), then
    the parent scheduler could correctly identify a backend-scope cache hit using information
    stored from the prior call node. This is left to future work.

    Parameters
    ----------
    expr: TaskExpression
        TaskExpression to be run by sub-scheduler.
    config
        A two-level python dict to configure the sub-scheduler.  (Will be used to initialize a
        :class:`Config` object via the `config_dict` kwarg.)
    load_modules
        List of modules that must be imported before starting the sub-scheduler. Before
        launching the sub-scheduler, the modules that define the user tasks must be imported.
    run_config
        Sub-scheduler run() kwargs. These are typically the local-scheduler run() kwargs that
        should also apply to the sub-scheduler such as `dryrun` and `cache` settings.

    Returns
    -------
    Dict[str, Any]
        Returns a dict result from running the sub-scheduler on `expr`.  The dict contains extra
        information pertaining to the sub-scheduler as follows:
        {
          'result': sub-scheduler evaluation of `expr`, present if successful
          'error': exception raised by sub-scheduler, present if error was encountered
          'dryrun': present if dryrun was requested and workflow can't complete

          'job_id': Job id of root Job in sub-scheduler,
                    present if parent_job_id given in run_config
          'call_hash': call_hash of root CallNode in sub-scheduler,
                       present if parent_job_id given in run_config

          'status': sub-scheduler final job status poll (list of str)
          'config': sub-scheduler config dict (useful for confirming exactly what config
            settings were used).
          'run_config': a dict of run configuration used by the sub-scheduler
        }

        parent_job_id is present in run_config, only results are return. Errors are raised.
    """
    sub_scheduler = Scheduler(config=Config(config_dict=config))
    sub_scheduler.load()

    # Import user modules.  The user is responsible for ensuring that the Executor that runs this
    # task has access to all these modules' code.  Additionally, any other user modules imported
    # by these modules must also be accessible.
    if load_modules:
        for module in load_modules:
            importlib.import_module(module)

    subrun_result: dict = {
        "config": sub_scheduler.config.get_config_dict(),
    }

    # Extend an existing Execution if the parent_job_id is given. Otherwise, start
    # a new Execution.
    if "parent_job_id" in run_config:
        result = sub_scheduler.extend_run(expr.eval(), **run_config)

        # If extending an Execution, supplement result value with additional
        # information such as job id and call_hash.
        if not isinstance(result, dict):
            raise AssertionError(f"Unknown scheduler result: {result}")
        subrun_result.update(result)

    else:
        result = sub_scheduler.run(expr.eval(), **run_config)
        subrun_result["result"] = result

    subrun_result.update(
        {
            "run_config": {
                "dryrun": sub_scheduler._dryrun,
                "cache": sub_scheduler._use_cache,
            },
            "status": sub_scheduler.get_job_status_report(),
        }
    )
    return subrun_result


@scheduler_task(
    namespace="redun",
    version="1",
    cache_scope=CacheScope.BACKEND,
    check_valid=CacheCheckValid.SHALLOW,
)
def subrun(
    scheduler: Scheduler,
    parent_job: Job,
    sexpr: SchedulerExpression,
    expr: Any,
    executor: str,
    config: Optional[Dict[str, Any]] = None,
    new_execution: bool = False,
    **task_options: dict,
) -> Promise:
    """
    Evaluates an expression `expr` in a sub-scheduler running within Executor `executor`.

    `executor` and optional `task_options` are used to configure the special redun task (
    _subrun_root_task) that starts the sub-scheduler. For example, you can configure the task
    with a batch executor to run the sub-scheduler on AWS Batch.

    `config`: To ease configuration management of the sub-scheduler, you can pass a `config`
    dict which contains configuration that would otherwise require a redun.ini file in the
    sub-scheduler environment. If you do not pass a config, the local scheduler's config will be
    forwarded to the sub-scheduler (replacing the local `config_dir` with "."). In practice,
    the sub-scheduler's `config_dir` should be less important as you probably want to log both
    local and sub-scheduler call graphs to a common database. You can also obtain a copy of the
    local scheduler's config and customize it as needed. Instantiate the scheduler directly instead
    of calling `redun run`.  Then access its config via
    `scheduler.py::get_scheduler_config_dict()`.

    Note on code packaging: The user is responsible for ensuring that the chosen Executor for
    invoking the sub-scheduler copies over all user-task scripts.  E.g. the local scheduler may
    be launched on local tasks defined in workflow1.py but subrun(executor="batch) is invoked on
    taskX defined in workflow2.py.  In this case, the user must ensure workflow2.py is copied to
    the batch node by placing it within the same directory tree as workflow1.py.

    The cache behavior of `subrun` is customized. Since recursive reductions are supposed to
    happen inside the subrun, we can't accept a single reduction. If the user wants full
    validity checking, that requires using single reductions, which means we have to actually
    start the sub-scheduler and let it handle the caching logic. In contrast, if a CSE or
    ultimate reduction is available, then we won't need to start the sub-scheduler.

    Most scheduler tasks don't accept the cache options `cache_scope` and `check_valid`,
    but `subrun` does, so we can forward them to the underlying `Task` that implements the subrun.
    Note that we also set the default `check_valid` value to be `CacheCheckValid.SHALLOW`, unlike
    its usual value.

    Parameters
    ----------
    expr : Any
        Expression to be run by sub-scheduler.
    executor : str
        Executor name for the special redun task that launches the sub-scheduler
        (_subrun_root_task). E.g. `batch` to launch the sub-scheduler in AWS Batch. Note
        that this is a config key in the  *local* scheduler's config.
    config : dict
        Optional sub-scheduler config dict. Must be a two-level dict that can be used to
        initialize a :class:`Config` object (see :method:`Config.get_config_dict()`).  If None or
        empty, the local Scheduler's config will be passed to the sub-scheduler and any values
        with local config_dir will be replaced with ".".  Do not include database credentials as
        they will be logged as clear text in the call graph.
    new_execution : bool
        If True, record the provenance of the evaluation of `expr` as a new Execution, otherwise
        extend the current Execution with new Jobs.
    **task_options : Any
        Task options for _subrun_root_task.  E.g. when running the sub-scheduler via Batch
        executor, you can pass ECS configuration (e.g. `memory`, `vcpus`, `batch_tags`).

    Returns
    -------
    Dict[str, Any]
        Returns a dict result of evaluating the _subrun_root_task
    """

    # If a config wasn't provided, forward the parent scheduler's backend config to the
    # sub-scheduler, replacing the local config_dir with "."
    if not config:
        config = scheduler.config.get_config_dict(replace_config_dir=".")

    # Collect a list of all modules that define user task.  This list will include the initial
    # task(s) to be subrun() as these are registered in the local task registry during subrun()
    # argument formulation.
    registry = get_task_registry()
    load_modules = set()
    for _task in registry:
        module_name = _task.load_module
        if module_name.startswith("redun.") and not module_name.startswith("redun.tests."):
            continue
        load_modules.add(module_name)

    # Prepare child task call for subrun.
    run_config: Dict[str, Any] = {
        "dryrun": scheduler._dryrun,
        "cache": scheduler._use_cache,
    }
    if not new_execution:
        run_config["parent_job_id"] = parent_job.id

    # Extract the cache options.
    all_options: Dict[str, Any] = {
        "cache_scope": CacheScope(
            sexpr.task_expr_options.get("cache_scope", subrun.get_task_option("cache_scope"))
        ),
        "check_valid": CacheCheckValid(
            sexpr.task_expr_options.get("check_valid", subrun.get_task_option("check_valid"))
        ),
        "allowed_cache_results": {CacheResult.CSE, CacheResult.ULTIMATE},
    }
    all_options.update(task_options)

    subrun_root_task_expr = _subrun_root_task.options(executor=executor, **all_options)(
        expr=quote(expr),
        config=config,
        load_modules=sorted(load_modules),
        run_config=run_config,
    )

    def log_banner(msg):
        scheduler.log("-" * 79)
        scheduler.log(msg)
        scheduler.log("-" * 79)

    def then(subrun_result):
        if "job_id" in subrun_result:
            # Create stub-job representing the job in the subscheduler.
            # The call_hash is needed to compute the right call_hash of parent_job.
            job = Job(
                root_task,
                root_task(quote(None)),
                id=subrun_result["job_id"],
                parent_job=parent_job,
                execution=scheduler._current_execution,
            )
            job.call_hash = subrun_result["call_hash"]

        # Echo sub-scheduler report.
        scheduler.log()
        log_banner(
            f"BEGIN: Sub-scheduler report (executor='{executor}'), {trim_string(repr(expr))}"
        )

        scheduler.log("Sub-scheduler config:")
        scheduler.log(pprint.pformat(subrun_result["config"]))

        scheduler.log()
        scheduler.log("Sub-scheduler run_config:")
        scheduler.log(pprint.pformat(subrun_result["run_config"]))

        scheduler.log()
        scheduler.log("Sub-scheduler final job status:")
        for report_line in subrun_result["status"]:
            scheduler.log(report_line)

        log_banner("END: Sub-scheduler report")
        scheduler.log()

        if "result" in subrun_result:
            # Return the user's result.
            return subrun_result["result"]

        elif "error" in subrun_result:
            # Reraise the error.
            raise subrun_result["error"]

        elif "dryrun" in subrun_result:
            # Return a promise that never resolves.
            return Promise()

    return scheduler.evaluate(subrun_root_task_expr, parent_job=parent_job).then(then)



import inspect
from typing import Any

import pytest
from sqlalchemy.orm import Session

from redun import Scheduler, apply_tags, task
from redun.backends.base import TagEntity
from redun.backends.db import Execution, Job, RedunBackendDb, Tag, Task, Value
from redun.tags import (
    ANY_VALUE,
    format_tag_key_value,
    format_tag_value,
    parse_tag_key_value,
    parse_tag_value,
)


@pytest.mark.parametrize(
    "text,value",
    [
        ("10", 10),
        ("1.2", 1.2),
        ("1.2e3", 1.2e3),
        ("true", True),
        ("false", False),
        ("null", None),
        ("", None),
        (" ", " "),
        ('""', ""),
        ('"10"', "10"),
        ('"true"', "true"),
        ("[1, 2, 3]", [1, 2, 3]),
        ('{"a": 1, "b": 2}', {"a": 1, "b": 2}),
        ("hello", "hello"),
        ("10a", "10a"),
        ("one two", "one two"),
        ("one [two]", "one [two]"),
        # Invalid tag values.
        ("[hello]", ValueError),
        ("[1, 2, 3", ValueError),
        ('"1, 2" 3', ValueError),
    ],
)
def test_parse_tag_value(text: str, value: Any) -> None:
    """
    Should be able to parse custom formats for tag values.
    """
    if inspect.isclass(value) and issubclass(value, Exception):
        with pytest.raises(value):
            parse_tag_value(text)
    else:
        assert parse_tag_value(text) == value


@pytest.mark.parametrize(
    "text,key_value",
    [
        ("key=10", ("key", 10)),
        ("project=acme", ("project", "acme")),
        ("project", ("project", ANY_VALUE)),
        ("equation=x=y", ("equation", "x=y")),
        ("key=", ("key", None)),
        ("a b=c d", ("a b", "c d")),
        ("=10", ValueError),
        ("", ValueError),
        ("10=10", ("10", 10)),
    ],
)
def test_parse_tag_key_value(text: str, key_value: Any) -> None:
    """
    Should be able to parse tag key value pairs.
    """
    if inspect.isclass(key_value) and issubclass(key_value, Exception):
        with pytest.raises(key_value):
            parse_tag_key_value(text, value_required=False)
    else:
        assert parse_tag_key_value(text, value_required=False) == key_value


def test_parse_tag_key_value_required() -> None:
    """
    Error should be raised when tag is missing a required value.
    """
    with pytest.raises(ValueError):
        parse_tag_key_value("project", value_required=True)


@pytest.mark.parametrize(
    "value,text",
    [
        (10, "10"),
        (True, "true"),
        (False, "false"),
        (None, "null"),
        ("", '""'),
        ("hello", "hello"),
        ("10a", "10a"),
        ("true", '"true"'),
        ("false", '"false"'),
        ("null", '"null"'),
        ("10", '"10"'),
        ("10.2", '"10.2"'),
        ("[10]", '"[10]"'),
        ('{"a": 1}', r'"{\"a\": 1}"'),
        (" hello ", '" hello "'),
    ],
)
def test_format_tag_value(value: Any, text: str) -> None:
    """
    Should be able to format tag values correctly.
    """
    assert format_tag_value(value) == text


@pytest.mark.parametrize(
    "key,value,text",
    [
        ("project", "acme", "project=acme"),
        ("cost", 10, "cost=10"),
        ("doc", "a234567890123", "doc=a234567..."),
        ("a234567890123", "a234567890123", "a234567...=a234567..."),
    ],
)
def test_format_tag_key_value(key: str, value: Any, text: str) -> None:
    """
    Should be able to format tag values correctly.
    """
    assert format_tag_key_value(key, value, max_length=10) == text


def test_value_tags(scheduler: Scheduler, session: Session) -> None:
    """
    We should be able to add tags to a Value.
    """
    input_tags = {"env": "prod", "project": "acme", "data": [1, "a", True]}

    @task(cache=False)
    def task1():
        x = 10
        return apply_tags(x, list(input_tags.items()))

    scheduler.run(task1())

    # Tags should be recorded.
    tags = session.query(Tag).filter(Tag.key.in_(input_tags.keys()))
    tag_data = {tag.key: tag.value for tag in tags}
    assert tag_data == {
        "env": "prod",
        "project": "acme",
        "data": [1, "a", True],
    }

    # They should have Value as an entity.
    assert isinstance(tags[0].entity, Value)
    assert tags[0].entity.value_parsed == 10

    # Running the same task again should not create more of the same tags.
    scheduler.run(task1())
    tags = session.query(Tag).filter(Tag.key.in_(input_tags.keys()))
    assert tags.count() == 3

    value = session.query(Value).filter(Value.value_hash == tags[0].entity_id).one()
    assert {tag.key for tag in value.tags} == {"env", "project", "data"}


def test_apply_all_tags(scheduler: Scheduler, session: Session) -> None:
    """
    We should be able to add tags to a Value, Job, and Execution.
    """

    @task
    def task1():
        x = 10
        return apply_tags(x, [("a", 1)], job_tags=[("b", 2)], execution_tags=[("c", 3)])

    scheduler.run(task1())

    # Tags should be recorded on value.
    job = session.query(Job).one()
    value = job.call_node.value
    assert {tag.key: tag.value for tag in value.tags} == {"a": 1}

    # Tags should be recorded on job.
    assert {tag.key: tag.value for tag in job.tags} == {"b": 2}

    # Tags should be recorded on execution.
    execution = session.query(Execution).one()
    assert {tag.key: tag.value for tag in execution.tags} == {"c": 3}


def test_task_tags(scheduler: Scheduler, session: Session) -> None:
    """
    We should be able to add tags to a Task.
    """

    @task(tags=[("env", "dev")])
    def task2():
        return 10

    scheduler.run(task2.options(tags=[("user", "alice")])())

    # Task tags are defined in the decorator.
    task_tag = session.query(Tag).filter(Tag.entity_id == task2.hash).one()
    assert task_tag.key == "env"
    assert task_tag.value == "dev"

    # They should have Task as an entity.
    assert isinstance(task_tag.entity, Task)
    assert task_tag.entity.hash == task2.hash

    task_row = session.query(Task).one()
    task_row.tags[0].key == "env"

    # Job tags are defined at call time.
    job_tag = session.query(Tag).filter(Tag.entity_type == TagEntity.Job).one()
    assert job_tag.key == "user"
    assert job_tag.value == "alice"
    assert isinstance(job_tag.entity, Job)
    assert job_tag.entity.task.name == "task2"

    job = session.query(Job).one()
    assert job.tags[0].key == "user"


def test_namespace_tags(scheduler: Scheduler, session: Session) -> None:
    """
    We should be able to create tags with namespaces.
    """
    tags = {"redun.project": "acme", "redun.sub.user": "alice", "env": "prod"}

    @task(cache=False)
    def task1():
        x = 10
        return apply_tags(x, list(tags.items()))

    scheduler.run(task1())

    # Tags should be recorded.
    tag1, tag2, tag3 = session.query(Tag).filter(Tag.key.in_(tags.keys())).order_by(Tag.key)

    assert tag1.base_key == "env"
    assert tag1.namespace == ""

    assert tag2.base_key == "project"
    assert tag2.namespace == "redun"

    assert tag3.base_key == "user"
    assert tag3.namespace == "redun.sub"


def test_execution_tags(scheduler: Scheduler, session: Session) -> None:
    """
    We should be able to create tags for Executions.
    """

    @task(cache=False)
    def task1():
        return 10

    scheduler.run(task1(), tags=[("project", "acme")])
    tags = session.query(Tag).all()
    execution = session.query(Execution).one()

    expected_tag_keys = {"project", "git_commit", "git_origin_url"}
    for tag in tags:
        assert tag.entity_id == execution.id
        assert tag.entity_type == TagEntity.Execution
        assert tag.key in expected_tag_keys
        if tag.key == "project":
            assert tag.value == "acme"


def test_tag_edit(backend: RedunBackendDb, session: Session) -> None:
    """
    We should be able to edit a tag.
    """
    # Record a Value (10).
    entity = 10
    entity_id = backend.record_value(entity)

    # Tag the Value with env=dev.
    [(tag_hash, _, _, _)] = backend.record_tags(TagEntity.Value, entity_id, [("env", "dev")])
    assert backend.get_tags([entity_id]) == {entity_id: {"env": ["dev"]}}

    # Update the tag with env=prod.
    [tag2] = backend.record_tags(TagEntity.Value, entity_id, [("env", "prod")], parents=[tag_hash])
    assert backend.get_tags([entity_id]) == {entity_id: {"env": ["prod"]}}

    # Ensure TagEdit edges are recorded.
    tag_row = session.query(Tag).filter_by(is_current=True).one()
    assert tag_row.parents[0].value == "dev"
    assert session.query(Tag.value).filter_by(is_current=True).one()[0] == "prod"

    # Update the tag with env=prod2 and add tag user=alice.
    backend.record_tags(
        TagEntity.Value, entity_id, [("env", "prod2"), ("user", "alice")], update=True
    )
    assert backend.get_tags([entity_id]) == {entity_id: {"env": ["prod2"], "user": ["alice"]}}

    # Rename the tag key.
    backend.update_tags(TagEntity.Value, entity_id, ["env"], [("environ", "production")])
    assert backend.get_tags([entity_id]) == {
        entity_id: {"environ": ["production"], "user": ["alice"]}
    }


def test_tag_edit_idempotent(backend: RedunBackendDb, session: Session) -> None:
    """
    Adding tags should be idempotent.
    """
    # Record a Value (10).
    entity = 10
    entity_id = backend.record_value(entity)

    [(tag_hash, _, _, _)] = backend.record_tags(TagEntity.Value, entity_id, [("key", "value")])
    backend.record_tags(
        TagEntity.Value,
        entity_id,
        [("key", "value2")],
        parents=[tag_hash],
    )
    backend.record_tags(
        TagEntity.Value,
        entity_id,
        [("key", "value2")],
        parents=[tag_hash],
    )
    assert backend.get_tags([entity_id]) == {entity_id: {"key": ["value2"]}}


def test_tag_merge(backend: RedunBackendDb, session: Session) -> None:
    """
    We should be able to merge multiple values together for a tag.
    """
    # Record a Value (10).
    entity = 10
    entity_id = backend.record_value(entity)

    # Tag the Value with env=dev.
    backend.record_tags(TagEntity.Value, entity_id, [("env", "dev")])
    assert backend.get_tags([entity_id]) == {entity_id: {"env": ["dev"]}}

    # In parallel, tag the Value with env=dev.
    backend.record_tags(TagEntity.Value, entity_id, [("env", "prod")])
    assert backend.get_tags([entity_id]) == {entity_id: {"env": ["dev", "prod"]}}

    # Merge tags.
    backend.record_tags(TagEntity.Value, entity_id, [("env", "prod")], update=True)
    assert backend.get_tags([entity_id]) == {entity_id: {"env": ["prod"]}}


def test_tag_delete(backend: RedunBackendDb, session: Session) -> None:
    """
    We should be able to merge multiple values together for a tag.
    """
    # Record a Value (10).
    entity = 10
    entity_id = backend.record_value(entity)

    # Tag the Value with env=dev.
    backend.record_tags(TagEntity.Value, entity_id, [("env", "dev"), ("env", "prod")])
    assert backend.get_tags([entity_id]) == {entity_id: {"env": ["dev", "prod"]}}

    # Delete tag.
    backend.delete_tags(entity_id, [("env", "dev")])
    assert backend.get_tags([entity_id])[entity_id] == {"env": ["prod"]}


def test_tag_new(backend: RedunBackendDb, session: Session) -> None:
    """
    We should be able to recreate a tag after its deleted.
    """
    # Record a Value (10).
    entity = 10
    entity_id = backend.record_value(entity)

    # Tag the Value with env=dev.
    backend.record_tags(TagEntity.Value, entity_id, [("env", "dev"), ("env", "prod")])
    assert backend.get_tags([entity_id]) == {entity_id: {"env": ["dev", "prod"]}}

    # Delete tag.
    backend.delete_tags(entity_id, [("env", "dev")])
    assert backend.get_tags([entity_id])[entity_id] == {"env": ["prod"]}

    # Trying to make the same tag without parents, will not bring the tag back.
    backend.record_tags(TagEntity.Value, entity_id, [("env", "dev"), ("env", "prod")])
    assert backend.get_tags([entity_id]) == {entity_id: {"env": ["prod"]}}

    # Using new=True, we will automatically create the tag as a leaf in the TagGraph.
    backend.record_tags(TagEntity.Value, entity_id, [("env", "dev"), ("env", "prod")], new=True)
    assert backend.get_tags([entity_id]) == {entity_id: {"env": ["dev", "prod"]}}

    # Delete the tag again.
    backend.delete_tags(entity_id, [("env", "dev")])
    assert backend.get_tags([entity_id])[entity_id] == {"env": ["prod"]}

    # Add tag back again.
    backend.record_tags(TagEntity.Value, entity_id, [("env", "dev"), ("env", "prod")], new=True)
    assert backend.get_tags([entity_id]) == {entity_id: {"env": ["dev", "prod"]}}



import os
import tempfile

from redun import File, Scheduler, task
from redun.backends.db import Execution
from redun.backends.db import File as RowFile
from redun.backends.db import Job, RedunBackendDb, Value


def test_record_job(scheduler: Scheduler) -> None:
    @task()
    def task1():
        return 10

    @task()
    def workflow(x):
        return task1()

    scheduler.run(workflow(1), exec_argv=["redun", "run", "workflow.py", "workflow", "--x", "1"])
    scheduler.run(workflow(2), exec_argv=["redun", "run", "workflow.py", "workflow", "--x", "2"])

    assert isinstance(scheduler.backend, RedunBackendDb)
    assert scheduler.backend.session
    jobs = scheduler.backend.session.query(Job).order_by(Job.start_time).all()
    assert len(jobs) == 4

    # Ensure the jobs were recorded.
    assert jobs[0].task.name == "workflow"
    assert jobs[1].task.name == "task1"
    assert jobs[0].task.source == "    def workflow(x):\n        return task1()\n"

    # Tasks should also point back to their jobs.
    assert len(jobs[0].task.jobs) == 2

    # The second pair of jobs are partially cached.
    assert not jobs[2].cached
    assert jobs[3].cached

    # Jobs should be connected into a tree.
    assert jobs[0].child_jobs[0].task.name == "task1"
    assert jobs[0].child_jobs[0].parent_job.task.name == "workflow"

    # CallNodes should be connected into a graph.
    assert jobs[0].call_node.child_edges[0].child_node.task.name == "task1"
    assert jobs[0].call_node.parents == []
    assert jobs[0].call_node.children[0].task.name == "task1"

    executions = scheduler.backend.session.query(Execution).all()
    assert executions[0].job.task.name == "workflow"

    assert executions[0].job.call_node.value_parsed == 10


def test_record_multicall(scheduler: Scheduler) -> None:
    """
    CallNodes can have repeat children and parents. The database should
    not apply a distinct to call_node.children or call_node.parents.
    """

    @task
    def id(x):
        return x

    @task
    def task1(x):
        return x

    @task
    def workflow():
        return [
            task1(10),
            task1(id(10)),
        ]

    scheduler.run(workflow())

    assert isinstance(scheduler.backend, RedunBackendDb)
    assert scheduler.backend.session
    jobs = scheduler.backend.session.query(Job).order_by(Job.start_time).all()

    # Let's get the top CallNode.
    call_node = jobs[0].call_node
    assert call_node.task.name == "workflow"

    # We should have three child CallNodes, even though two are the same CallNode.
    assert len(call_node.children) == 3
    assert len(set(call_node.children)) == 2

    child_node = call_node.children[0]
    assert len(child_node.parents) == 2
    assert len(set(child_node.parents)) == 1


def test_record_file(scheduler: Scheduler) -> None:
    """
    We should record files that are pass through tasks.
    """

    @task()
    def task1(path):
        file = File(os.path.join(path, "myfile"))
        file.write("hello")
        return file

    @task()
    def task2(file):
        return file.read()

    @task()
    def main(path):
        file = task1(path)
        return [task2(file), file]

    with tempfile.TemporaryDirectory() as tmpdir:
        scheduler.run(main(tmpdir))

    # Our file should have been recorded by its type.
    assert isinstance(scheduler.backend, RedunBackendDb)
    assert scheduler.backend.session
    my_file = scheduler.backend.session.query(Value).filter_by(type="redun.File").one()

    # We should have a record of which tasks returned this File.
    # TODO: Could I have a record that the file came from main() as well?
    assert {result.task.name for result in my_file.results} == {"task1"}

    # We also have a record of who accepted this File as input.
    assert my_file.arguments[0].arg_position == 0
    assert my_file.arguments[0].call_node.task.name == "task2"

    # Ensure subvalues were detected.
    # my_file was also part of a list that was returned from main.
    assert my_file.parents[0].type == "builtins.list"
    assert my_file.parents[0].results[0].task.name == "main"

    # Ensure ee can get File information like path.
    assert my_file.file.path
    path = my_file.file.path

    # We can also search the logs starting with file path.
    files = scheduler.backend.session.query(RowFile).filter_by(path=path).all()
    my_file2 = files[0].value
    assert {result.task.name for result in my_file2.results} == {"task1"}



import os
import pickle
import sys
from collections import defaultdict, namedtuple
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum
from typing import Any, DefaultDict, Dict, Generator, Generic, NamedTuple, Optional, TypeVar

import pytest

from redun.tests.utils import use_tempdir
from redun.utils import (
    MultiMap,
    PreviewClass,
    PreviewUnpicklable,
    add_import_path,
    clear_import_paths,
    get_import_paths,
    iter_nested_value,
    map_nested_value,
    merge_dicts,
    pickle_dumps,
    pickle_preview,
)

T = TypeVar("T", int, float, str)


@dataclass
class Data(Generic[T]):
    x: T
    y: Dict[str, Any]
    z: Optional["Data"]  # noqa: F821

    def echo(self, string: str) -> str:
        return string


@dataclass
class CustomData1(Data[float]):
    x: float


CustomData2 = Data[int]


def test_iter_nested_value() -> None:
    value = {
        "a": [1, [2, 3]],
        "b": {4.0, True, None},
        "c": ("7", "888", 9),
        "d": CustomData1(x=1.0, y={"foo": "bar"}, z=CustomData2(x=42, y={}, z=None)),
    }
    assert set(iter_nested_value(value)) == {
        "a",
        "b",
        "c",
        "d",
        1,
        2,
        3,
        4.0,
        True,
        None,
        "7",
        "888",
        9,
        1.0,
        "foo",
        "bar",
        42,
    }


def test_map_nested_value() -> None:
    """
    map_nested_value should map functions on supported container types:
    list, dict, set, tuple, nested tuple.

    map_nested_value should not map functions to other types currently
    """

    def func(value) -> str:
        return str(value) + "a"

    # Nested values are supported for namedtuple, but not defaultdict currently
    my_namedtuple = namedtuple("my_namedtuple", "a")
    test_defaultdict: DefaultDict[str, int] = defaultdict(int)
    test_defaultdict["a"]
    test_namedtuple = my_namedtuple(a=1)
    test_dataclass = CustomData1(x=1.0, y={"foo": "bar"}, z=CustomData2(x=42, y={}, z=None))

    value = [
        [1, [2]],
        {"a": 1, "b": {"c": 2}},
        (1, (2,)),
        {1, (2,)},
        test_namedtuple,
        test_defaultdict,
        test_dataclass,
    ]

    assert map_nested_value(func, value) == [
        ["1a", ["2a"]],
        {"aa": "1a", "ba": {"ca": "2a"}},
        ("1a", ("2a",)),
        {("2a",), "1a"},
        my_namedtuple(a="1a"),
        "defaultdict(<class 'int'>, {'a': 0})a",
        CustomData1(
            x="1.0a", y={"fooa": "bara"}, z=CustomData2(x="42a", y={}, z="Nonea")  # type: ignore
        ),
    ]


@pytest.mark.parametrize(
    "dicts,expected",
    [
        # Base cases.
        ([], {}),
        ([{}], {}),
        ([{"a": 1}], {"a": 1}),
        # Combine keys.
        (
            [
                {"a": 1, "b": 2},
                {"b": 3, "c": 4},
            ],
            {"a": 1, "b": 3, "c": 4},
        ),
        (
            [{"a": 1, "b": 2}, {"b": 3, "c": 4}, {"d": 5}],
            {"a": 1, "b": 3, "c": 4, "d": 5},
        ),
        (
            [
                {"a": {"a1": 1}},
                {"a": {"a2": 2}},
            ],
            {"a": {"a1": 1, "a2": 2}},
        ),
        # Mix dict vs non-dict.
        (
            [
                {"a": {"b": 1}},
                {"a": 2},
                {"a": 3},
            ],
            {"a": 3},
        ),
    ],
)
def test_merge_dicts(dicts, expected) -> None:
    """
    Should be able to merge nested dictionaries together.
    """
    assert merge_dicts(dicts) == expected


def test_import_paths() -> None:
    """
    We should be able to prepend new import paths, retrieve them, clear them.
    """
    # New paths should prepend.
    original_paths = list(sys.path)
    add_import_path("mypath")
    add_import_path("mypath2")
    assert sys.path == ["mypath2", "mypath"] + original_paths

    # We should be able to retrieve just the redun paths.
    assert get_import_paths() == ["mypath2", "mypath"]

    # We should be able to clear redun paths.
    clear_import_paths()
    assert get_import_paths() == []


# Class to pickle for test functions.
class User:
    def __init__(self, name: str, age: int, friend: "User" = None):
        self.name = name
        self.age = age
        self.friend = friend


@contextmanager
def hide_class(name) -> Generator:
    user_class = globals()[name]
    del globals()[name]
    yield
    globals()[name] = user_class


def test_pickle_preview_class() -> None:
    """Tests pickle preview when pickled class is not imported"""

    y = [2, 3]
    x = [1, y, User("Alice", 30, User("Bob", 31))]
    data = pickle_dumps(x)

    with hide_class("User"):
        # AttributeError should raise since User and myfunc can't be found.
        with pytest.raises(AttributeError):
            pickle.loads(data)

        # However, pickle_preview() still works by using stub classes.
        x2 = pickle_preview(data)
    assert (
        str(x2) == "[1, [2, 3], User(name='Alice', age=30,"
        " friend=User(name='Bob', age=31, friend=None))]"
    )

    assert isinstance(x2[2], PreviewClass)
    assert x2[2].name == "Alice"
    assert x2[2].age == 30
    assert isinstance(x2[2].friend, PreviewClass)
    assert x2[2].friend.name == "Bob"
    assert x2[2].friend.age == 31

    with pytest.raises(AttributeError):
        assert x2[2].invalid == "whoops"


def test_pickle_preview_dataclass() -> None:
    """Tests pickle preview when pickled dataclass is not imported"""
    x = 1
    y = {"2": {"3": "4"}}
    z = CustomData1(x, y, None)
    obj = CustomData1(x, y, z)

    data = pickle_dumps(obj)

    with hide_class("CustomData1"):
        # AttributeError should raise since Data can't be found.
        with pytest.raises(AttributeError):
            pickle.loads(data)

        # However, pickle_preview() still works by using stub classes.
        obj2 = pickle_preview(data)
    assert (
        str(obj2)
        == "CustomData1(x=1, y={'2': {'3': '4'}}, z=CustomData1(x=1, y={'2': {'3': '4'}}, z=None))"
    )

    assert isinstance(obj2, PreviewClass)
    assert obj2.x == 1
    assert obj2.y == {"2": {"3": "4"}}
    assert isinstance(obj2.z, PreviewClass)
    assert obj2.z.x == 1
    assert obj2.z.y == {"2": {"3": "4"}}
    assert obj2.z.z is None

    with pytest.raises(AttributeError):
        assert obj2.invalid == "whoops"


@use_tempdir
def test_pickle_preview_module() -> None:
    """Tests pickle preview when pickled module is not importable"""

    # External module containing class definition.
    with open("the_class.py", "w") as temp:
        temp.write(
            """
class ConfusedUser:
    def __init__(self, name: str, confusion: int):
        self.name = name
        self.confusion = confusion
        """
        )

    # Create a class instance and pickle it.
    sys.path.append(".")
    from the_class import ConfusedUser

    x = ConfusedUser("Robin", 9000)
    data = pickle_dumps(x)

    # Delete the module and unimport it.
    os.remove("the_class.py")
    del sys.modules["the_class"]

    with pytest.raises(ModuleNotFoundError):
        pickle.loads(data)

    x2 = pickle_preview(data)

    assert isinstance(x2, PreviewClass)
    assert x2.name == "Robin"
    assert x2.confusion == 9000


def test_pickle_unpicklable() -> None:
    """Tests pickle preview of un-unpicklable objects"""

    data = b"This is not a valid pickle file!"

    with pytest.raises(pickle.UnpicklingError):
        pickle.loads(data)

    x2 = pickle_preview(data)
    assert isinstance(x2, PreviewUnpicklable)
    assert isinstance(x2.error, pickle.UnpicklingError)
    assert str(x2.error) == "pickle data was truncated"


class Color(Enum):
    RED = 1
    GREEN = 2
    BLUE = 3


def test_pickle_preview_enum() -> None:
    """
    Pickle preview should instantiate Enums.
    """
    data = pickle_dumps(Color.RED)

    with hide_class("Color"):
        with pytest.raises(AttributeError):
            pickle.loads(data)

        obj = pickle_preview(data)

    assert repr(obj) == "Color(1)"
    assert obj[0] == 1


class UserTuple(NamedTuple):
    name: str
    age: int


def test_pickle_preview_namedtuple() -> None:
    """
    Pickle preview should instantiate Enums.
    """
    data = pickle_dumps(UserTuple("Alice", 30))

    with hide_class("UserTuple"):
        with pytest.raises(AttributeError):
            pickle.loads(data)

        obj = pickle_preview(data)

    assert repr(obj) == "UserTuple('Alice', 30)"
    assert obj[0] == "Alice"
    assert obj[1] == 30


def test_multimap() -> None:
    """
    Test that a MultiMap supports its operations.
    """
    mmap: MultiMap = MultiMap([["a", 1], ["a", 2], ["b", "X"], ["c", "Y"]])

    assert len(mmap) == 4

    assert mmap["a"] == [1, 2]
    assert mmap["b"] == ["X"]
    assert mmap.get("c") == ["Y"]
    assert mmap.get("d") == []
    assert mmap.get("e", [True]) == [True]

    assert "c" in mmap
    assert "d" not in mmap
    assert mmap.has_item("a", 2)
    assert not mmap.has_item("a", 3)
    assert not mmap.has_item("d", 3)

    assert dict(iter(mmap)) == {"a": 2, "b": "X", "c": "Y"}
    assert sorted(mmap.keys()) == ["a", "b", "c"]
    assert set(mmap.values()) == {1, 2, "X", "Y"}
    assert list(mmap) == [("a", 1), ("a", 2), ("b", "X"), ("c", "Y")]
    assert mmap.as_dict() == {
        "a": [1, 2],
        "b": ["X"],
        "c": ["Y"],
    }

    mmap2 = MultiMap(mmap)
    assert mmap == mmap2
    assert mmap != MultiMap([[10, 20], [30, 40]])
    assert mmap == {
        "a": [1, 2],
        "b": ["X"],
        "c": ["Y"],
    }
    assert mmap == list(mmap)

    # Add key-value pairs.
    mmap.add("d", 7)
    mmap.add("d", 8)
    mmap.add("e", 9)
    assert mmap.get("d") == [7, 8]
    assert len(mmap) == 7



import sys
from typing import Any, Optional

_current_namespace = ""


def namespace(_namespace=None):
    """
    Set the current task namespace.

    Use None to unset the current namespace.
    """
    global _current_namespace
    _current_namespace = _namespace or ""


def get_current_namespace():
    """
    Returns the current task namespace during import.
    """
    return _current_namespace


def compute_namespace(obj: Any, namespace: Optional[str] = None):
    """Compute the namespace for the provided object.

    Precedence:
    - Explicit namespace provided (note: an empty string is a valid explicit value)
    - Infer it from a `redun_namespace` variable in the same module as func
    - The current namespace, configured with `set_current_namespace`

    WARNING: This computation interacts with the global "current namespace" (see above),
    so results may be context dependent.
    """

    # Determine task namespace.
    if namespace is None:
        namespace = getattr(sys.modules[obj.__module__], "redun_namespace", None)

    if namespace is not None:
        return namespace
    return get_current_namespace()



from unittest.mock import Mock

import pytest

from redun.promise import Promise, wait_promises


def test_resolve() -> None:
    """
    Promise constructor should resolve values.
    """
    promise: Promise[int] = Promise(lambda resolve, reject: resolve(10))
    assert promise.value == 10


def test_reject() -> None:
    """
    Promise constructor should reject exceptions.
    """
    error = ValueError("boom")
    promise: Promise[int] = Promise(lambda resolve, reject: reject(error))
    assert promise.error == error


def test_do_resolve() -> None:
    """
    Promises should resolve values after instantiation.
    """
    promise: Promise[int] = Promise()
    promise.do_resolve(10)
    assert promise.value == 10


def test_do_reject() -> None:
    """
    Promises should reject exceptions after instantiation.
    """
    error = ValueError("boom")
    promise: Promise[int] = Promise()
    promise.do_reject(error)
    assert promise.error == error


def test_resolve_first() -> None:
    """
    First resolution counts.
    """
    error = ValueError("boom")
    promise: Promise[int] = Promise()
    promise.do_resolve(10)
    promise.do_resolve(20)
    promise.do_reject(error)
    assert promise.value == 10


def test_resolved_has_no_error() -> None:
    """
    Once resolved, promise is marked as not rejected.
    """
    error = ValueError("boom")
    promise: Promise[int] = Promise()
    promise.do_resolve(10)
    promise.do_reject(error)
    assert promise.value == 10
    with pytest.raises(ValueError):
        promise.error


def test_reject_first() -> None:
    """
    First rejection counts.
    """
    error = ValueError("boom")
    error2 = ValueError("boom2")
    promise: Promise[int] = Promise()
    promise.do_reject(error)
    promise.do_reject(error2)
    promise.do_resolve(10)
    promise.do_resolve(20)
    assert promise.error == error


def test_then_resolve() -> None:
    """
    Promise resolutions should propagate through then().
    """
    mock = Mock()
    promise: Promise[int] = Promise()
    promise.then(mock)
    promise.do_resolve(10)
    mock.assert_called_with(10)


def test_then_reject() -> None:
    """
    Promise rejections should propagate through then().
    """

    def fail(error):
        pass

    mock = Mock(fail)
    promise: Promise[int] = Promise()
    promise.then(None, mock)

    error = ValueError("boom")
    promise.do_reject(error)
    mock.assert_called_with(error)


def test_resolve_then() -> None:
    """
    Already resolved promises should still propagate through then().
    """
    promise: Promise[int] = Promise()
    promise.do_resolve(10)

    mock = Mock()
    promise.then(mock)
    mock.assert_called_with(10)


def test_reject_then() -> None:
    """
    Already rejected promises should still propagate through then().
    """
    promise: Promise[int] = Promise()
    error = ValueError("boom")
    promise.do_reject(error)

    mock = Mock()
    promise.then(None, mock)
    mock.assert_called_with(error)


def test_resolve_multiple_then() -> None:
    """
    Resolutions should fan-out to multiple `then()` calls.
    """
    promise: Promise[int] = Promise()

    mock = Mock()
    promise.then(mock)
    mock2 = Mock()
    promise.then(mock2)

    promise.do_resolve(10)
    mock.assert_called_with(10)
    mock2.assert_called_with(10)


def test_reject_multiple_then() -> None:
    """
    Rejections should fan-out through multiple `then()` calls.
    """
    promise: Promise[int] = Promise()

    mock = Mock()
    promise.then(None, mock)
    mock2 = Mock()
    promise.then(None, mock2)

    error = ValueError("boom")
    promise.do_reject(error)
    mock.assert_called_with(error)
    mock2.assert_called_with(error)


def test_chain_resolve() -> None:
    """
    Resolutions should propagate through chained `then()` calls.
    """
    promise: Promise[int] = Promise()

    mock = Mock()
    promise.then(lambda x: x + 1).then(lambda x: x + 2).then(mock)

    promise.do_resolve(10)
    mock.assert_called_with(13)


def test_chain_reject() -> None:
    """
    Rejections should propagate through chained `then()` calls.
    """
    promise: Promise[int] = Promise()

    mock = Mock()
    promise.then(lambda x: x + 1).then(None, mock)

    error = ValueError("boom")
    promise.do_reject(error)
    mock.assert_called_with(error)


def test_chain_null() -> None:
    """
    Resolutions should propagate through empty `then()` calls.
    """
    promise: Promise[int] = Promise()

    mock = Mock()
    promise.then().then(mock)

    promise.do_resolve(10)
    mock.assert_called_with(10)


def test_nest_resolve() -> None:
    """
    Resolutions of nested promises should propagate.
    """
    promise: Promise[int] = Promise()

    mock = Mock()
    promise.then(lambda x: Promise(lambda resolve, reject: resolve(x + 1))).then(mock)

    promise.do_resolve(10)
    mock.assert_called_with(11)


def test_nest_reject() -> None:
    """
    Rejections of nested promises should propagate.
    """
    promise: Promise[int] = Promise()

    error = ValueError("boom")
    mock = Mock()
    promise.then(lambda x: Promise(lambda resolve, reject: reject(error))).catch(mock)

    promise.do_resolve(10)
    mock.assert_called_with(error)


def test_raise() -> None:
    """
    Raising an exception should reject the Promise.
    """
    error = ValueError("boom")

    def then(result: int) -> None:
        raise error

    def fail(error: Exception) -> int:
        return 100

    def final(result: int) -> None:
        pass

    promise: Promise[int] = Promise()

    mock = Mock(side_effect=fail)
    mock2 = Mock(side_effect=final)
    promise.then(then).catch(mock).then(mock2)

    promise.do_resolve(10)
    mock.assert_called_with(error)
    mock2.assert_called_with(100)


def test_all() -> None:
    """
    Promise.all() should resolve when all child promises resolve.
    """
    promise1: Promise[int] = Promise()
    promise2: Promise[int] = Promise()

    promises = Promise.all([promise1, promise2])

    mock = Mock()
    promises.then(mock)

    promise1.do_resolve(1)
    promise2.do_resolve(2)
    mock.assert_called_with([1, 2])


def test_all_first() -> None:
    """
    Promise.all() should resolve even when all child promises resolve first.
    """
    promise1: Promise[int] = Promise()
    promise2: Promise[int] = Promise()
    promise1.do_resolve(1)
    promise2.do_resolve(2)

    promises = Promise.all([promise1, promise2])

    mock = Mock()
    promises.then(mock)

    mock.assert_called_with([1, 2])


def test_all_reject() -> None:
    """
    Promise.all() should reject when any child promise rejects.
    """
    promise1: Promise[int] = Promise()
    promise2: Promise[int] = Promise()

    promises = Promise.all([promise1, promise2])

    mock = Mock()
    promises.then(None, mock)

    error = ValueError("boom")
    promise1.do_resolve(1)
    promise2.do_reject(error)
    mock.assert_called_with(error)


def test_all_reject_first() -> None:
    """
    Promise.all() should reject even when a child promise rejects first.
    """
    error = ValueError("boom")
    promise1: Promise[int] = Promise()
    promise2: Promise[int] = Promise()
    promise1.do_resolve(1)
    promise2.do_reject(error)

    promises = Promise.all([promise1, promise2])

    mock = Mock()
    promises.then(None, mock)

    mock.assert_called_with(error)


def test_wait_promises() -> None:
    """
    wait_promises() should wait for all subpromises regardless of success or failure.
    """

    def good():
        return Promise(lambda resolve, reject: resolve(10))

    def bad():
        return Promise(lambda resolve, reject: reject(ValueError("boom")))

    top = wait_promises([good(), good(), good()])
    promises = top.value
    assert [promise.value for promise in promises] == [10, 10, 10]

    top = wait_promises([good(), bad(), good()])
    promises = top.value
    assert promises[0].value == 10
    assert isinstance(promises[1].error, ValueError)
    assert promises[2].value == 10

    a: Promise[int] = Promise()
    top = wait_promises([good(), bad(), a])
    a.do_resolve(20)
    promises = top.value
    assert promises[0].value == 10
    assert isinstance(promises[1].error, ValueError)
    assert promises[2].value == 20



import os
from typing import Tuple

from redun.file import File


class ValueStore:
    """
    Stores redun Values on filesystem (local or remote).

    The value store can be configured within the `[backend]` section using
    the `value_store_path` and `value_store_min_size` variables.

    Values in this store are organized similarly to git objects:

    <value_store_path>/ab/cdef... for value with hash abcdef...

    """

    def __init__(self, root_path: str, use_subdir: bool = True):
        self.root_path = root_path
        self.use_subdir = use_subdir

    def get_value_path(self, value_hash: str) -> str:
        """
        Return the path for a value.
        """
        if self.use_subdir:
            suffix = value_hash[:2] + "/" + value_hash[2:]
        else:
            suffix = value_hash

        return os.path.join(self.root_path, suffix)

    def has(self, value_hash: str) -> bool:
        """
        Return True if value is in store.
        """
        return File(self.get_value_path(value_hash)).exists()

    def put(self, value_hash: str, data: bytes) -> None:
        """
        Store Value data.
        """
        if self.has(value_hash):
            return

        with File(self.get_value_path(value_hash)).open("wb") as out:
            out.write(data)

    def get(self, value_hash: str) -> Tuple[bytes, bool]:
        """
        Retrieve Value data.
        """
        file = File(self.get_value_path(value_hash))
        try:
            with file.open("rb") as infile:
                return infile.read(), True
        except FileNotFoundError:
            return b"", False

    def size(self, value_hash: str) -> int:
        """
        Returns the size in bytes of a Value.
        """
        try:
            return File(self.get_value_path(value_hash)).size()
        except FileNotFoundError:
            return -1
